"""Implementations of the built-in Scenic classes.

Defines the 3 Scenic classes `Point`, `OrientedPoint`, and `Object`, and associated
helper code (notably their base class `Constructible`, which implements the handling of
property definitions and :ref:`specifier resolution`).

.. warning::

    In :ref:`2D compatibility mode`, these classes are overwritten with 2D analogs. While
    we make an effort to map imports to the correct class, this only works if imports
    use the form ``import scenic.core.object_types as object_types`` followed by accessing
    ``object_types.Object``. If you instead use ``from scenic.core.object_types import Object``,
    you may get the wrong class.

"""

from abc import ABC, abstractmethod
import collections
import math
import random
import typing
import warnings

import numpy as np
import shapely
import shapely.affinity
import trimesh

from scenic.core.distributions import (
    FunctionDistribution,
    MultiplexerDistribution,
    RandomControlFlowError,
    Samplable,
    distributionFunction,
    distributionMethod,
    needsSampling,
    supportInterval,
    toDistribution,
)
from scenic.core.errors import InvalidScenarioError, SpecifierError
from scenic.core.geometry import (
    averageVectors,
    hypot,
    max,
    min,
    normalizeAngle,
    pointIsInCone,
)
from scenic.core.lazy_eval import (
    LazilyEvaluable,
    isLazy,
    needsLazyEvaluation,
    valueInContext,
)
from scenic.core.regions import (
    BoxRegion,
    CircularRegion,
    EmptyRegion,
    IntersectionRegion,
    MeshSurfaceRegion,
    MeshVolumeRegion,
    PolygonalRegion,
    Region,
    SectorRegion,
    SpheroidRegion,
    ViewRegion,
    convertToFootprint,
)
from scenic.core.serialization import dumpAsScenicCode
from scenic.core.shapes import BoxShape, MeshShape, Shape
from scenic.core.specifiers import ModifyingSpecifier, PropertyDefault, Specifier
from scenic.core.type_support import (
    toHeading,
    toOrientation,
    toScalar,
    toType,
    toVector,
    underlyingType,
)
from scenic.core.utils import DefaultIdentityDict, cached_method, cached_property
from scenic.core.vectors import (
    Orientation,
    Vector,
    alwaysGlobalOrientation,
    globalOrientation,
)
from scenic.core.visibility import canSee

## Types
#: Type alias for an interval (a pair of floats).
Interval = typing.Tuple[float, float]
#: Type alias for limits on dimensions (a triple of intervals).
DimensionLimits = typing.Tuple[Interval, Interval, Interval]

## Abstract base class


class Constructible(Samplable):
    """Abstract base class for Scenic objects.

    Scenic objects, which are constructed using specifiers, are implemented
    internally as instances of ordinary Python classes. This abstract class
    implements the procedure to resolve specifiers and determine values for
    the properties of an object, as well as several common methods supported
    by objects.

    .. warning::

        This class is an implementation detail, and none of its methods should be
        called directly from a Scenic program.
    """

    _dynamicProperties = {}

    def __init_subclass__(cls):
        super().__init_subclass__()

        if "_defaults" in cls.__dict__:
            # This class is being unpickled by value; the pickled class already was
            # transformed by __init_subclass__, so we skip it now.
            return

        # Identify cached properties/methods which will need to be cleared
        # each time step during dynamic simulations.
        clearers = {}
        for attr, value in cls.__dict__.items():
            if isinstance(value, property):
                value = value.fget
            if clearer := getattr(value, "_scenic_cache_clearer", None):
                clearers[attr] = clearer
        for sc in cls.__mro__:
            if sclearers := getattr(sc, "_cache_clearers", None):
                for attr, clearer in sclearers.items():
                    if attr not in clearers:
                        clearers[attr] = clearer
        cls._cache_clearers = clearers

        # Find all defaults provided by the class or its superclasses
        allDefs = collections.defaultdict(list)

        for sc in cls.__mro__:
            if issubclass(sc, Constructible) and hasattr(sc, "_scenic_properties"):
                for prop, value in sc._scenic_properties.items():
                    allDefs[prop].append(PropertyDefault.forValue(value))

        # Resolve conflicting defaults and gather dynamic properties
        resolvedDefs = {}
        dyns = []
        finals = []
        dynFinals = {}
        for prop, defs in allDefs.items():
            primary, rest = defs[0], defs[1:]
            spec = primary.resolveFor(prop, rest)
            resolvedDefs[prop] = spec

            if isDynamic := any(defn.isDynamic for defn in defs):
                dyns.append(prop)
            if primary.isFinal:
                finals.append(prop)
                if isDynamic:
                    dynFinals[prop] = primary.value
        cls._defaults = resolvedDefs
        cls._finalProperties = frozenset(finals)

        # Determine types of dynamic properties
        dynTypes = {}
        defaultValues = None  # compute only if necessary
        for prop in dyns:
            ty = super(cls, cls)._dynamicProperties.get(prop)
            if not ty:
                # First time this property has been defined; get the type of
                # its default value.
                if not defaultValues:
                    # N.B. Here we evaluate the default value expressions, which is
                    # risky since global state like the workspace may not have been set
                    # up yet. For this reason we only compute default values when they
                    # are actually needed; a better solution would be to have syntax for
                    # annotating the types of dynamic properties.
                    defaultValues, _ = cls._resolveSpecifiers(())
                ty = underlyingType(defaultValues[prop])
            dynTypes[prop] = ty
        cls._dynamicProperties = dynTypes
        cls._simulatorProvidedProperties = {
            prop: val
            for prop, val in cls._dynamicProperties.items()
            if prop not in cls._finalProperties
        }

        # Extract order in which to recompute dynamic final properties each time step
        if defaultValues:
            recomputers = {}
            for prop in defaultValues:  # order is that from specifier resolution
                if prop in dynFinals:
                    recomputers[prop] = dynFinals[prop]
            cls._dynamicFinalProperties = recomputers
        else:
            # No new dynamic properties: just inherit the order from the superclass
            pass

    def __new__(cls, *args, _internal=False, **kwargs):
        if not _internal:
            # Catch users trying to instantiate a Scenic class like a Python class
            raise InvalidScenarioError('Scenic classes must be instantiated with "new"')
        return super().__new__(cls)

    def __getnewargs_ex__(self):
        return ((), dict(_internal=True))

    def __init__(self, properties, constProps=frozenset(), _internal=False):
        for prop, value in properties.items():
            assert not needsLazyEvaluation(value), (prop, value)
            object.__setattr__(self, prop, value)
        super().__init__(properties.values())
        self.properties = tuple(sorted(properties.keys()))
        self._propertiesSet = set(self.properties)
        self._constProps = constProps

    @classmethod
    def _withProperties(cls, properties, constProps=None):
        """Create an instance with the given property values.

        Values of unspecified properties are determined by specifier resolution
        as usual.
        """
        specs = []
        for prop, val in properties.items():
            specs.append(Specifier(f"<internal({prop})>", {prop: 1}, {prop: val}))
        return cls._withSpecifiers(specs, constProps=constProps, register=False)

    @classmethod
    def _with(cls, **properties):
        # Shorthand form of _withProperties
        return cls._withProperties(properties)

    @classmethod
    def _withSpecifiers(cls, specifiers, constProps=None, register=True):
        """Create an instance from the given specifiers."""
        # Resolve specifiers
        newspecs = cls._prepareSpecifiers(specifiers)
        properties, consts = cls._resolveSpecifiers(newspecs)
        if constProps is None:
            constProps = consts

        # Catch properties which would conflict with ordinary attributes
        for prop in properties:
            if hasattr(cls, prop):
                raise SpecifierError(
                    f"Property {prop} would overwrite an attribute with the same name."
                )

        # Create the object
        obj = cls(properties, constProps=constProps, _internal=True)

        # Possibly register this object
        if register:
            obj._register()

        return obj

    @classmethod
    def _prepareSpecifiers(cls, specifiers):
        # This is a hook for subclasses to modify the specifier list.
        return specifiers

    @classmethod
    def _resolveSpecifiers(cls, specifiers, defaults=None, overriding=False):
        specifiers = list(specifiers)

        # Declare properties dictionary which maps properties to the specifier
        # that will specify that property.
        properties = dict()

        # Declare modifying dictionary, which maps properties to a specifier
        # that will modify that property.
        modifying = dict()

        # Dictionary mapping properties set so far to the priority with which they have
        # been set.
        priorities = dict()

        # Extract default property values dictionary and set of final properties,
        # unless defaults is overriden.
        if defaults is None:
            defaults = cls._defaults

        finals = cls._finalProperties

        # Check for incompatible specifier combinations
        specifiers_count = collections.Counter(spec.name for spec in specifiers)

        for spec in specifiers_count:
            if specifiers_count[spec] > 1:
                raise SpecifierError(f"Cannot use {spec} specifier to modify itself.")

        # Split the specifiers into two groups, normal and modifying. Normal specifiers set all relevant properties
        # first. Then modifying specifiers can modify or set additional properties
        normal_specifiers = [
            spec for spec in specifiers if not isinstance(spec, ModifyingSpecifier)
        ]
        modifying_specifiers = [
            spec for spec in specifiers if isinstance(spec, ModifyingSpecifier)
        ]

        # For each property specified by a normal specifier:
        #   - If not in properties specified, properties[p] = specifier
        #   - Otherwise, if property specified, check if specifier's priority is higher. If so, replace it with specifier

        # Priorties are inversed: A lower priority number means semantically that it has a higher priority level
        for spec in normal_specifiers:
            assert isinstance(spec, Specifier), (name, spec)

            # Iterate over each property.
            for prop in spec.priorities:
                # Check if this is a final property that has been specified.
                if prop in finals:
                    raise SpecifierError(
                        f'property "{prop}" cannot be directly specified'
                    )

                if prop in properties:
                    # This property already exists. Check that it has not already been specified
                    # at equal priority level. Then if it was previously specified at a lower priority
                    # level, override it with the value that this specifier sets.
                    if spec.priorities[prop] == priorities[prop]:
                        raise SpecifierError(
                            f'property "{prop}" specified twice with the same priority'
                        )
                    if spec.priorities[prop] < priorities[prop]:
                        properties[prop] = spec
                        priorities[prop] = spec.priorities[prop]
                else:
                    # This property has not already been specified, so we should initialize it.
                    properties[prop] = spec
                    priorities[prop] = spec.priorities[prop]

        # If a modifying specifier specifies the property with a higher priority,
        # set the object's property to be specified by the modifying specifier. Otherwise,
        # if the property exists and has already been specified at a higher or equal priority,
        # then the resulting value is modified by the modifying specifier.

        # If the property is not yet being specified, the modifying specifier will
        # act as a normal specifier for that property.
        for spec in modifying_specifiers:
            for prop in spec.priorities:
                # Now we check if the propert has already been specified
                if prop in properties:
                    # This property has already been specified, so we should either modify
                    # it or specify it.
                    if spec.priorities[prop] < priorities[prop]:
                        # Higher priority level, so it specifies
                        properties[prop] = spec
                        priorities[prop] = spec.priorities[prop]
                    elif prop in spec.modifiable_props:
                        # This specifer can modify this prop, so we set it to do so after
                        # first checking it has not already been modified.
                        if prop in modifying:
                            raise SpecifierError(
                                f'property "{prop}" of {name} modified twice.'
                            )

                        modifying[prop] = spec
                else:
                    # This property has not been specified, so we should specify it.
                    properties[prop] = spec
                    priorities[prop] = spec.priorities[prop]

        # Add any default specifiers needed
        _defaultedProperties = set()
        for prop, default_spec in defaults.items():
            if prop not in priorities:
                specifiers.append(default_spec)
                properties[prop] = default_spec
                _defaultedProperties.add(prop)

        # Create the actual_props dictionary, which maps each specifier to a set of properties
        # it is actually specifying or modifying.
        actual_props = {spec: [] for spec in specifiers}
        for prop in properties:
            # Extract the specifier that is specifying this prop and add it to the
            # specifier's entry in actual_props
            specifying_spec = properties[prop]
            actual_props[specifying_spec].append(prop)

            # If a specifier modifies this property, add this prop to the specifiers
            # actual_props list.
            if prop in modifying:
                modifying_spec = modifying[prop]
                actual_props[modifying_spec].append(prop)

        # Create an inversed modifying dictionary that specifiers to the properties they
        # are modifying.
        modifying_inv = {spec: prop for prop, spec in modifying.items()}

        # Topologically sort specifiers. Specifiers become vertices and the properties
        # those specifiers depend on become the in-edges of each vertex. The specifiers
        # are then sorted topologically according to this graph.
        order = []
        for spec in specifiers:
            spec._dfs_state = 0

        def dfs(spec):
            if spec._dfs_state == 2:  # finished processing this specifier
                return
            elif spec._dfs_state == 1:  # specifier is being processed
                raise SpecifierError(f"specifier {spec.name} depends on itself")
            spec._dfs_state = 1

            # Recurse on dependencies
            for dep in spec.requiredProperties:
                child = modifying.get(dep)
                if not child:
                    child = properties.get(dep)

                if child is None:
                    raise SpecifierError(
                        f"property {dep} required by "
                        f"specifier {spec} is not specified"
                    )
                else:
                    dfs(child)

            # If this is a modifying specifier, recurse on the specifier
            # that specifies the property being modified.
            if spec in modifying_inv:
                specifying_spec = properties[modifying_inv[spec]]
                dfs(specifying_spec)

            order.append(spec)
            spec._dfs_state = 2

        for spec in specifiers:
            dfs(spec)
        assert len(order) == len(specifiers)
        for spec in specifiers:
            del spec._dfs_state

        context = LazilyEvaluable.makeContext()
        for spec in order:
            specifiedValues = spec.getValuesFor(context)
            for prop in actual_props[spec]:
                assert not hasattr(context, prop) or prop in modifying, (prop, spec)
                value = toDistribution(specifiedValues[prop])
                cls._specify(context, prop, value)
        properties = LazilyEvaluable.getContextValues(context)

        constProps = frozenset(
            {prop for prop in _defaultedProperties if not needsSampling(properties[prop])}
        )
        return properties, constProps

    def _recomputeDynamicFinals(self):
        # Evaluate default value expression for each dynamic final property
        # and assign the obtained value
        for prop, recomputer in self._dynamicFinalProperties.items():
            rawVal = recomputer(self)
            value = valueInContext(rawVal, self)
            self._specify(self, prop, value)

    @classmethod
    def _specify(cls, context, prop, value):
        # Normalize types of some built-in properties
        if prop in (
            "position",
            "velocity",
            "cameraOffset",
            "positionStdDev",
            "orientationStdDev",
        ):
            value = toVector(value, f'"{prop}" of {cls.__name__} not a vector')
        elif prop in (
            "width",
            "length",
            "visibleDistance",
            "viewAngle",
            "speed",
            "angularSpeed",
            "yaw",
            "pitch",
            "roll",
            "mutationScale",
        ):
            value = toScalar(value, f'"{prop}" of {cls.__name__} not a scalar')

        if prop in ["yaw", "pitch", "roll"]:
            value = normalizeAngle(value)

        if prop == "parentOrientation":
            value = toOrientation(value)

        if prop == "regionContainedIn":
            # 2D regions can't contain objects, so we automatically use their footprint.
            value = convertToFootprint(value)

        if prop == "color" and value is not None and not isLazy(value):
            if any(not (0 <= v <= 1) for v in value):
                raise ValueError(
                    "Color property contains value not between 0 and 1 (inclusive)."
                )

            if not 3 <= len(value) <= 4:
                raise ValueError(f"Color property has incorrect length {len(value)}.")

        object.__setattr__(context, prop, value)

    def _register(self):
        import scenic.syntax.veneer as veneer  # TODO improve?

        veneer.registerInstance(self)

    def _override(self, specifiers):
        assert not needsSampling(self)
        # Validate properties being overridden and gather their old values
        oldVals = {}
        for spec in specifiers:
            for prop in spec.priorities:
                if prop in self._dynamicProperties:
                    raise SpecifierError(f'cannot override dynamic property "{prop}"')
                if prop not in self._propertiesSet:
                    raise SpecifierError(f'object has no property "{prop}" to override')
                oldVals[prop] = getattr(self, prop)

        # Perform specifier resolution to find the new values of all properties
        defs = {
            prop: Specifier("OverrideDefault", {prop: -1}, {prop: getattr(self, prop)})
            for prop in self.properties
        }
        newprops, _ = self._resolveSpecifiers(specifiers, defaults=defs)

        # Apply the new values
        for prop, val in newprops.items():
            object.__setattr__(self, prop, val)

        # If we assigned a new dynamic behavior, it might need to be started.
        if behavior := newprops["behavior"]:
            behavior._assignTo(self)

        return oldVals

    def _revert(self, oldVals):
        for prop, val in oldVals.items():
            object.__setattr__(self, prop, val)

    def sampleGiven(self, value):
        if not needsSampling(self):
            return self
        props = {prop: value[getattr(self, prop)] for prop in self.properties}
        return type(self)(props, constProps=self._constProps, _internal=True)

    def _allProperties(self):
        return {prop: getattr(self, prop) for prop in self.properties}

    def _copyWith(self, **overrides):
        """Copy this object, possibly overriding some of its properties."""
        # Copy all properties except for final values, which will retain their default values
        props = {
            prop: val
            for prop, val in self._allProperties().items()
            if prop not in self._finalProperties
        }
        props.update(overrides)
        constProps = self._constProps.difference(overrides)
        return self._withProperties(props, constProps=constProps)

    def _clearCaches(self):
        for clearer in self._cache_clearers.values():
            clearer(self)

    def dumpAsScenicCode(self, stream, skipConstProperties=True):
        stream.write(f"new {self.__class__.__name__}")
        first = True
        for prop in self.properties:
            if skipConstProperties and prop in self._constProps:
                continue
            if prop in self._finalProperties:
                continue
            if prop == "position":
                spec = "at"
            else:
                spec = f"with {prop}"
            if first:
                stream.write(" ")
                first = False
            else:
                stream.write(",\n    ")
            stream.write(f"{spec} ")
            dumpAsScenicCode(getattr(self, prop), stream)

    def __str__(self):
        if hasattr(self, "properties") and "name" in self._propertiesSet:
            return self.name
        else:
            return f"unnamed {self.__class__.__name__}"

    def __repr__(self):
        if hasattr(self, "properties"):
            allProps = {prop: getattr(self, prop) for prop in self.properties}
        else:
            allProps = "<under construction>"
        return f"{type(self).__name__}({allProps})"


## Mutators


class Mutator:
    """An object controlling how the :keyword:`mutate` statement affects an `Object`.

    A `Mutator` can be assigned to the ``mutator`` property of an `Object` to
    control the effect of the :keyword:`mutate` statement. When mutation is enabled
    for such an object using that statement, the mutator's `appliedTo` method
    is called to compute a mutated version. The `appliedTo` method can also decide
    whether to apply mutators inherited from superclasses.
    """

    def appliedTo(self, obj):
        """Return a mutated version of the given object. Implemented by subclasses.

        The mutator may inspect the ``mutationScale`` attribute of the given object
        to scale its effect according to the scale given in ``mutate O by S``.

        Returns:
            A pair consisting of the mutated version of the object together with a
            Boolean indicating whether the mutator inherited from the superclass
            (if any) should also be applied.
        """
        raise NotImplementedError


class PositionMutator(Mutator):
    """Mutator adding Gaussian noise to ``position``. Used by `Point`.

    Attributes:
        stddevs (tuple[float,float,float]): standard deviation of noise for each dimension (x,y,z).
    """

    def __init__(self, stddevs):
        self.stddevs = tuple(stddevs)

    def appliedTo(self, obj):
        noise = Vector(
            random.gauss(0, self.stddevs[0] * obj.mutationScale),
            random.gauss(0, self.stddevs[1] * obj.mutationScale),
            random.gauss(0, self.stddevs[2] * obj.mutationScale),
        )
        obj.position += noise
        return (obj, True)  # allow further mutation

    def __eq__(self, other):
        if type(other) is not type(self):
            return NotImplemented
        return other.stddevs == self.stddevs

    def __hash__(self):
        return hash(self.stddevs)


class OrientationMutator(Mutator):
    """Mutator adding Gaussian noise to ``yaw``, ``pitch``, and ``roll``. Used by `OrientedPoint`.

    Attributes:
        stddevs (tuple[float,float,float]): standard deviation of noise for each angle (yaw, pitch, roll).
    """

    def __init__(self, stddevs):
        self.stddevs = tuple(stddevs)

    def appliedTo(self, obj):
        obj.yaw += random.gauss(0, self.stddevs[0] * obj.mutationScale)
        obj.pitch += random.gauss(0, self.stddevs[1] * obj.mutationScale)
        obj.roll += random.gauss(0, self.stddevs[2] * obj.mutationScale)

        return (obj, True)  # allow further mutation

    def __eq__(self, other):
        if type(other) is not type(self):
            return NotImplemented
        return other.stddevs == self.stddevs

    def __hash__(self):
        return hash(self.stddevs)


## Point


class Point(Constructible):
    """The Scenic base class ``Point``.

    The default mutator for `Point` adds Gaussian noise to ``position`` with
    a standard deviation given by the ``positionStdDev`` property.

    Properties:
        position (`Vector`; dynamic): Position of the point. Default value is the origin (0,0,0).
        width (float): Default value 0 (only provided for compatibility with
          operators that expect an `Object`).
        length (float): Default value 0.
        height (float): Default value 0.
        baseOffset (`Vector`): Only provided for compatibility with the `on` specifier.
          Default value is (0,0,0).
        contactTolerance (float): Only provided for compatibility with the specifiers
          that expect an `Object`. Default value 0.
        onDirection (`Vector`): The direction used to determine where to place
          this `Point` on a region, when using the modifying :keyword:`on` specifier.
          See the :sampref:`on {region}` page for more details. Default value is None,
          indicating the direction will be inferred from the region this object is being placed on.
        visibleDistance (float): Distance used to determine the visible range of this object.
          Default value 50.
        viewRayDensity (float): By default determines the number of rays used during visibility checks.
          This value is the density of rays per degree of visible range in one dimension. The total
          number of rays sent will be this value squared per square degree of this object's view angles.
          This value determines the default value for ``viewRayCount``, so if ``viewRayCount`` is overwritten
          this value is ignored. Default value 5.
        viewRayCount (None | tuple[float, float]): The total number of horizontal and vertical view angles
          to be sent, or None if this value should be computed automatically. Default value ``None``.
        viewRayDistanceScaling (bool): Whether or not the number of rays should scale with the distance to the
          object. Ignored if ``viewRayCount`` is passed. Default value ``False``.
        mutationScale (float): Overall scale of mutations, as set by the
          :keyword:`mutate` statement. Default value 0 (mutations disabled).
        positionStdDev (tuple[float, float, float]): Standard deviation of Gaussian noise
          for each dimension (x,y,z) to be added to this object's :prop:`position`
          when mutation is enabled with scale 1. Default value (1,1,0), mutating only the x,y values
          of the point.
    """

    _scenic_properties = {
        "position": PropertyDefault((), {"dynamic"}, lambda self: Vector(0.0, 0.0, 0.0)),
        "width": 0.0,
        "length": 0.0,
        "height": 0.0,
        "baseOffset": Vector(0, 0, 0),
        "contactTolerance": 0,
        "onDirection": None,
        # This property is defined in OrientedPoint, but we provide a default value
        # for Points for implementation convenience.
        "viewAngles": (math.tau, math.pi),
        "visibleDistance": 50,
        "viewRayDensity": 5,
        "viewRayCount": None,
        "viewRayDistanceScaling": False,
        "mutationScale": 0,
        "mutator": PropertyDefault(
            ("positionStdDev",),
            {"additive"},
            lambda self: PositionMutator((self.positionStdDev)),
        ),
        "positionStdDev": (1, 1, 0),
        # This property is defined in Object, but we provide a default empty value
        # for Points for implementation convenience.
        "regionContainedIn": None,
        # These properties are used internally to store entities that must be able to
        # or must be unable to observe (view) this entity.
        "_observingEntity": None,
        "_nonObservingEntity": None,
    }

    @cached_property
    def visibleRegion(self):
        """The :term:`visible region` of this object.

        The visible region of a `Point` is a sphere centered at its ``position`` with
        radius ``visibleDistance``.
        """
        dimensions = (self.visibleDistance, self.visibleDistance, self.visibleDistance)
        return SpheroidRegion(position=self.position, dimensions=dimensions)

    @cached_method
    def canSee(self, other, occludingObjects=tuple(), debug=False) -> bool:
        """Whether or not this `Point` can see ``other``.

        Args:
            other: A `Point`, `OrientedPoint`, or `Object` to check
              for visibility.
            occludingObjects: A list of objects that can occlude visibility.
        """
        return canSee(
            position=self.position,
            orientation=None,
            visibleDistance=self.visibleDistance,
            viewAngles=(math.tau, math.pi),
            rayCount=self.viewRayCount,
            rayDensity=self.viewRayDensity,
            distanceScaling=self.viewRayDistanceScaling,
            target=other,
            occludingObjects=occludingObjects,
            debug=debug,
        )

    @cached_property
    def corners(self):
        return (self.position,)

    def toVector(self) -> Vector:
        return self.position

    def sampleGiven(self, value):
        sample = super().sampleGiven(value)
        if value[self.mutationScale] != 0:
            for mutator in self.mutator:
                if mutator is None:
                    continue
                sample, proceed = mutator.appliedTo(sample)
                if not proceed:
                    break
            sample._recomputeDynamicFinals()
        return sample

    # Points automatically convert to Vectors when needed
    def __getattr__(self, attr):
        if hasattr(Vector, attr):
            return getattr(self.toVector(), attr)
        else:
            self.__getattribute__(attr)


## OrientedPoint


class OrientedPoint(Point):
    """The Scenic class ``OrientedPoint``.

    The default mutator for `OrientedPoint` adds Gaussian noise to ``yaw``, ``pitch``
    and ``roll``, using the three standard deviations (for yaw/pitch/roll respectively)
    given by the  ``orientationStdDev`` property. It then also applies the mutator for `Point`.
    By default the standard deviations for ``pitch`` and ``roll`` are zero so that, by
    default, only ``yaw`` is mutated.

    Properties:
        yaw (float; dynamic): Yaw of the `OrientedPoint` in radians in the local coordinate system
          provided by :prop:`parentOrientation`. Default value 0.
        pitch (float; dynamic): Pitch of the `OrientedPoint` in radians in the local coordinate system
          provided by :prop:`parentOrientation`. Default value 0.
        roll (float; dynamic): Roll of the `OrientedPoint` in radians in the local coordinate system
          provided by :prop:`parentOrientation`. Default value 0.
        parentOrientation (`Orientation`): The local coordinate system that the `OrientedPoint`'s
          :prop:`yaw`, :prop:`pitch`, and :prop:`roll` are interpreted in. Default
          value is the global coordinate system, where an object is flat in the XY plane,
          facing North.
        orientation (`Orientation`; dynamic; final): The orientation of the `OrientedPoint` relative
          to the global coordinate system. Derived from the :prop:`yaw`, :prop:`pitch`,
          :prop:`roll`, and :prop:`parentOrientation` of this `OrientedPoint` and non-overridable.
        heading (float; dynamic; final): Yaw value of this `OrientedPoint` in the global coordinate
          system. Derived from :prop:`orientation` and non-overridable.
        viewAngles (tuple[float,float]): Horizontal and vertical view angles of this `OrientedPoint`
          in radians. Horizontal view angle can be up to 2π and vertical view angle can be
          up to π. Values greater than these will be truncated. Default value is (2π, π)
        orientationStdDev (tuple[float,float,float]): Standard deviation of Gaussian noise to add to this
          object's Euler angles (yaw, pitch, roll) when mutation is enabled with scale 1.
          Default value (5°, 0, 0), mutating only the :prop:`yaw` of this `OrientedPoint`.
    """

    _scenic_properties = {
        "yaw": PropertyDefault((), {"dynamic"}, lambda self: 0),
        "pitch": PropertyDefault((), {"dynamic"}, lambda self: 0),
        "roll": PropertyDefault((), {"dynamic"}, lambda self: 0),
        "parentOrientation": globalOrientation,
        "orientation": PropertyDefault(
            {"yaw", "pitch", "roll", "parentOrientation"},
            {"dynamic", "final"},
            lambda self: (
                self.parentOrientation
                * Orientation.fromEuler(self.yaw, self.pitch, self.roll)
            ),
        ),
        # Heading is equal to orientation.yaw, which is equal to self.yaw if this OrientedPoint's
        # parentOrientation is the global orientation. Defined this way to simplify the value for pruning
        # purposes if possible.
        "heading": PropertyDefault(
            {"orientation"},
            {"dynamic", "final"},
            lambda self: (
                self.yaw
                if alwaysGlobalOrientation(self.parentOrientation)
                else self.orientation.yaw
            ),
        ),
        "viewAngle": math.tau,  # Primarily for backwards compatibility. Set viewAngles instead.
        "viewAngles": PropertyDefault(
            ("viewAngle",), set(), lambda self: (self.viewAngle, math.pi)
        ),
        "mutator": PropertyDefault(
            {"orientationStdDev"},
            {"additive"},
            lambda self: OrientationMutator(self.orientationStdDev),
        ),
        "headingStdDev": math.radians(
            5
        ),  # Primarily for backwards compatibility. Set orientationStdDev instead.
        "orientationStdDev": PropertyDefault(
            ("headingStdDev",), set(), lambda self: (self.headingStdDev, 0, 0)
        ),
    }

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        if self.viewAngles[0] > math.tau or self.viewAngles[1] > math.pi:
            warnings.warn(
                "ViewAngles can not have values greater than (math.tau, math.pi). Truncating values..."
            )
            self.viewAngles = (
                min(self.viewAngles[0], math.tau),
                min(self.viewAngles[1], math.pi),
            )

    @cached_property
    def visibleRegion(self):
        """The :term:`visible region` of this object.

        The visible region of an `OrientedPoint` restricts that of `Point` (a sphere with
        radius :prop:`visibleDistance`) based on the value of :prop:`viewAngles`. In
        general, it is a capped rectangular pyramid subtending an angle of
        :scenic:`viewAngles[0]` horizontally and :scenic:`viewAngles[1]` vertically, as
        long as those angles are less than π/2; larger angles yield various kinds of
        wrap-around regions. See `ViewRegion` for details.
        """
        return ViewRegion(
            visibleDistance=self.visibleDistance,
            viewAngles=self.viewAngles,
            position=self.position,
            rotation=self.orientation,
        )

    @cached_method
    def canSee(self, other, occludingObjects=tuple(), debug=False) -> bool:
        """Whether or not this `OrientedPoint` can see ``other``.

        Args:
            other: A `Point`, `OrientedPoint`, or `Object` to check
              for visibility.
            occludingObjects: A list of objects that can occlude visibility.
        """
        return canSee(
            position=self.position,
            orientation=self.orientation,
            visibleDistance=self.visibleDistance,
            viewAngles=self.viewAngles,
            rayCount=self.viewRayCount,
            rayDensity=self.viewRayDensity,
            distanceScaling=self.viewRayDistanceScaling,
            target=other,
            occludingObjects=occludingObjects,
            debug=debug,
        )

    def relativize(self, vec):
        pos = self.relativePosition(vec)
        return OrientedPoint._with(position=pos, parentOrientation=self.orientation)

    def relativePosition(self, vec):
        return self.position.offsetLocally(self.orientation, vec)

    def distancePast(self, vec):
        """Distance past a given point, assuming we've been moving in a straight line."""
        diff = self.position - vec
        return diff.rotatedBy(-self.heading).y

    def toHeading(self) -> float:
        return self.heading

    def toOrientation(self) -> Orientation:
        return self.orientation


## Object


class Object(OrientedPoint):
    """The Scenic class ``Object``.

    This is the default base class for Scenic classes.

    Properties:
        width (float): Width of the object, i.e. extent along its X axis.
          Default value of 1 inherited from the object's :prop:`shape`.
        length (float): Length of the object, i.e. extent along its Y axis.
          Default value of 1 inherited from the object's :prop:`shape`.
        height (float): Height of the object, i.e. extent along its Z axis.
          Default value of 1 inherited from the object's :prop:`shape`.
        shape (`Shape`): The shape of the object, which must be an instance of `Shape`.
          The default shape is a box, with default unit dimensions.
        allowCollisions (bool): Whether the object is allowed to intersect
          other objects. Default value ``False``.
        regionContainedIn (`Region` or ``None``): A `Region` the object is
          required to be contained in. If ``None``, the object need only be
          contained in the scenario's :term:`workspace`.
        baseOffset (`Vector`): An offset from the :prop:`position` of the Object
          to the base of the object, used by the `on` specifier. Default value
          is :scenic:`(0, 0, -self.height/2)`, placing the base of the Object at the bottom
          center of the Object's bounding box.
        contactTolerance (float): The maximum distance this object can be away from a
          surface to be considered on the surface. Objects are placed at half this
          distance away from a point when the `on` specifier or a directional specifier
          like `left of {Object}` is used. Default value 1e-4.
        sideComponentThresholds (`DimensionLimits`): Used to determine the
          various sides of an object (when using the default implementation).
          The three interior 2-tuples represent the maximum and minimum bounds
          for each dimension's (x,y,z) surface. See `defaultSideSurface` for details.
          Default value :scenic:`((-0.5, 0.5), (-0.5, 0.5), (-0.5, 0.5))`.
        cameraOffset (`Vector`): Position of the camera for the :keyword:`can see`
          operator, relative to the object's :prop:`position`. Default :scenic:`(0, 0, 0)`.
        requireVisible (bool): Whether the object is required to be visible
          from the ``ego`` object. Default value ``False``.
        occluding (bool): Whether or not this object can occlude other objects. Default
          value ``True``.
        showVisibleRegion (bool): Whether or not to display the visible region in the
          Scenic internal visualizer.
        color (tuple[float, float, float, float] or tuple[float, float, float] or `None`):
          An optional color (with optional alpha) property that is used by the internal
          visualizer, or possibly simulators. All values should be between 0 and 1.
          Default value ``None``
        velocity (`Vector`; *dynamic*): Velocity in dynamic simulations. Default value is
          the velocity determined by :prop:`speed` and :prop:`orientation`.
        speed (float; dynamic): Speed in dynamic simulations. Default value 0.
        angularVelocity (`Vector`; *dynamic*):
        angularSpeed (float; dynamic): Angular speed in dynamic simulations. Default
          value 0.
        behavior: Behavior for dynamic agents, if any (see :ref:`dynamics`). Default
          value ``None``.
        lastActions: Tuple of :term:`actions` taken by this agent in the last time step
          (an empty tuple if the object is not an agent or this is the first time step).
    """

    _scenic_properties = {
        "width": PropertyDefault(("shape",), {}, lambda self: self.shape.width),
        "length": PropertyDefault(("shape",), {}, lambda self: self.shape.length),
        "height": PropertyDefault(("shape",), {}, lambda self: self.shape.height),
        "shape": BoxShape(),
        "allowCollisions": False,
        "regionContainedIn": None,
        "baseOffset": PropertyDefault(
            ("height",), {}, lambda self: Vector(0, 0, -self.height / 2)
        ),
        "contactTolerance": 1e-4,
        "sideComponentThresholds": ((-0.5, 0.5), (-0.5, 0.5), (-0.5, 0.5)),
        "cameraOffset": Vector(0, 0, 0),
        "requireVisible": False,
        "occluding": True,
        "showVisibleRegion": False,
        "color": None,
        "velocity": PropertyDefault((), {"dynamic"}, lambda self: Vector(0, 0, 0)),
        "speed": PropertyDefault((), {"dynamic"}, lambda self: 0),
        "angularVelocity": PropertyDefault((), {"dynamic"}, lambda self: Vector(0, 0, 0)),
        "angularSpeed": PropertyDefault((), {"dynamic"}, lambda self: 0),
        "behavior": None,
        "lastActions": tuple(),
        # weakref to scenario which created this object, for internal use
        "_parentScenario": None,
    }

    def __new__(cls, *args, **kwargs):
        obj = super().__new__(cls, *args, **kwargs)
        # The _dynamicProxy attribute stores a mutable copy of the object used during
        # simulations, intercepting all attribute accesses to the original object;
        # we set this attribute very early to prevent problems during unpickling.
        object.__setattr__(obj, "_dynamicProxy", obj)
        return obj

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.hw = hw = self.width / 2
        self.hl = hl = self.length / 2
        self.hh = hh = self.height / 2
        self.radius = hypot(hw, hl, hh)  # circumcircle; for collision detection

        self._relations = []

    @classmethod
    def _specify(cls, context, prop, value):
        # Normalize types of some built-in properties
        if prop == "behavior" and value != None:
            import scenic.syntax.veneer as veneer  # TODO improve?

            value = toType(
                value, veneer.Behavior, f'"behavior" of {cls.__name__} not a behavior'
            )
        super()._specify(context, prop, value)

    def _register(self):
        import scenic.syntax.veneer as veneer  # TODO improve?

        veneer.registerObject(self)

    def __getattribute__(self, name):
        proxy = object.__getattribute__(self, "_dynamicProxy")
        return object.__getattribute__(proxy, name)

    def __setattr__(self, name, value):
        proxy = object.__getattribute__(self, "_dynamicProxy")
        object.__setattr__(proxy, name, value)

    def __delattr__(self, name):
        proxy = object.__getattribute__(self, "_dynamicProxy")
        object.__delattr__(proxy, name)

    def startDynamicSimulation(self):
        """Hook called when the object is created in a dynamic simulation.

        Does nothing by default; provided for objects to do simulator-specific
        initialization as needed.

        .. versionchanged:: 3.0

            This method is called on objects created in the middle of dynamic
            simulations, not only objects present in the initial scene.
        """
        pass

    @cached_method
    def containsPoint(self, point):
        """Whether or not the space this object occupies contains a point"""
        return self.occupiedSpace.containsPoint(point)

    @cached_method
    def distanceTo(self, point):
        """The minimal distance from the space this object occupies to a given point"""
        return self.occupiedSpace.distanceTo(point)

    @cached_method
    def intersects(self, other):
        """Whether or not this object intersects another object or region"""
        ## Type Checking ##
        if not isinstance(other, (Object, Region)):
            raise TypeError(
                f"Cannot compute intersection of Scenic Object with {type(other)}."
            )

        ## Heuristic Fast Paths ##
        # For two objects that are boxes and flat, we can take a fast route
        if self._isPlanarBox and (isinstance(other, Object) and other._isPlanarBox):
            if abs(self.position.z - other.position.z) > (self.height + other.height) / 2:
                return False

            self_poly = self._boundingPolygon
            other_poly = other._boundingPolygon
            return self_poly.intersects(other_poly)

        # For an object that is a box and flat with a polygonal region, we can
        # also take a fast route.
        if self._isPlanarBox and (
            isinstance(other, PolygonalRegion)
            and abs(self.position.z - other.z) <= self.height / 2
        ):
            return self._boundingPolygon.intersects(other.polygons)

        ## Default Case
        # Extract other's occupied space if it's an object
        if isinstance(other, Object):
            other_occupied_space = other.occupiedSpace
        else:
            other_occupied_space = other

        if isLazy(self.occupiedSpace) or isLazy(other_occupied_space):
            raise RandomControlFlowError(
                "Cannot compute intersection between Objects with non-fixed values."
            )

        return self.occupiedSpace.intersects(other_occupied_space)

    @cached_property
    def left(self):
        return self.relativize(Vector(-self.hw, 0))

    @cached_property
    def right(self):
        return self.relativize(Vector(self.hw, 0))

    @cached_property
    def front(self):
        return self.relativize(Vector(0, self.hl))

    @cached_property
    def back(self):
        return self.relativize(Vector(0, -self.hl))

    @cached_property
    def top(self):
        return self.relativize(Vector(0, 0, self.hh))

    @cached_property
    def bottom(self):
        return self.relativize(Vector(0, 0, -self.hh))

    @cached_property
    def frontLeft(self):
        return self.relativize(Vector(-self.hw, self.hl))

    @cached_property
    def frontRight(self):
        return self.relativize(Vector(self.hw, self.hl))

    @cached_property
    def backLeft(self):
        return self.relativize(Vector(-self.hw, -self.hl))

    @cached_property
    def backRight(self):
        return self.relativize(Vector(self.hw, -self.hl))

    @cached_property
    def topFrontLeft(self):
        return self.relativize(Vector(-self.hw, self.hl, self.hh))

    @cached_property
    def topFrontRight(self):
        return self.relativize(Vector(self.hw, self.hl, self.hh))

    @cached_property
    def topBackLeft(self):
        return self.relativize(Vector(-self.hw, -self.hl, self.hh))

    @cached_property
    def topBackRight(self):
        return self.relativize(Vector(self.hw, -self.hl, self.hh))

    @cached_property
    def bottomFrontLeft(self):
        return self.relativize(Vector(-self.hw, self.hl, -self.hh))

    @cached_property
    def bottomFrontRight(self):
        return self.relativize(Vector(self.hw, self.hl, -self.hh))

    @cached_property
    def bottomBackLeft(self):
        return self.relativize(Vector(-self.hw, -self.hl, -self.hh))

    @cached_property
    def bottomBackRight(self):
        return self.relativize(Vector(self.hw, -self.hl, -self.hh))

    @cached_property
    def visibleRegion(self):
        """The :term:`visible region` of this object.

        The visible region of an `Object` is the same as that of an `OrientedPoint` (see
        `OrientedPoint.visibleRegion`) except that it is offset by the value of
        :prop:`cameraOffset` (which is the zero vector by default).
        """
        true_position = self.position.offsetLocally(self.orientation, self.cameraOffset)
        return ViewRegion(
            visibleDistance=self.visibleDistance,
            viewAngles=self.viewAngles,
            position=true_position,
            rotation=self.orientation,
        )

    @cached_method
    def canSee(self, other, occludingObjects=tuple(), debug=False) -> bool:
        """Whether or not this `Object` can see ``other``.

        Args:
            other: A `Point`, `OrientedPoint`, or `Object` to check
              for visibility.
            occludingObjects: A list of objects that can occlude visibility.
        """
        true_position = self.position.offsetLocally(self.orientation, self.cameraOffset)
        return canSee(
            position=true_position,
            orientation=self.orientation,
            visibleDistance=self.visibleDistance,
            viewAngles=self.viewAngles,
            rayCount=self.viewRayCount,
            rayDensity=self.viewRayDensity,
            distanceScaling=self.viewRayDistanceScaling,
            target=other,
            occludingObjects=occludingObjects,
            debug=debug,
        )

    @cached_property
    def corners(self):
        """A tuple containing the corners of this object's bounding box"""
        hw, hl, hh = self.hw, self.hl, self.hh
        return (
            self.relativePosition(Vector(hw, hl, hh)),
            self.relativePosition(Vector(-hw, hl, hh)),
            self.relativePosition(Vector(-hw, -hl, hh)),
            self.relativePosition(Vector(hw, -hl, hh)),
            self.relativePosition(Vector(hw, hl, -hh)),
            self.relativePosition(Vector(-hw, hl, -hh)),
            self.relativePosition(Vector(-hw, -hl, -hh)),
            self.relativePosition(Vector(hw, -hl, -hh)),
        )

    @cached_property
    def _corners2D(self):
        hw, hl = self.hw, self.hl
        # Note: 2D show method assumes cyclic order of vertices
        return (
            self.relativePosition(Vector(hw, hl)),
            self.relativePosition(Vector(-hw, hl)),
            self.relativePosition(Vector(-hw, -hl)),
            self.relativePosition(Vector(hw, -hl)),
        )

    @cached_property
    def occupiedSpace(self):
        """A region representing the space this object occupies"""
        shape = self.shape
        return MeshVolumeRegion(
            mesh=shape.mesh,
            dimensions=(self.width, self.length, self.height),
            position=self.position,
            rotation=self.orientation,
            centerMesh=False,
            _internal=True,
            _isConvex=shape.isConvex,
        )

    @property
    def _isConvex(self):
        """Whether this object's shape is convex"""
        return self.shape.isConvex

    @property
    def _hasStaticBounds(self):
        deps = (
            self.position,
            self.orientation,
            self.shape,
            self.width,
            self.length,
            self.height,
        )
        return not any(needsSampling(v) for v in deps)

    @cached_property
    def boundingBox(self):
        """A region representing this object's bounding box"""
        return MeshVolumeRegion(self.occupiedSpace.mesh.bounding_box, centerMesh=False)

    @cached_property
    def inradius(self):
        """A lower bound on the inradius of this object"""

        # Define a helper function that computes the support of the inradius,
        # given the sub supports.
        def inradiusSupport(width_s, length_s, height_s, shape_s):
            # Unpack the dimension supports (and ignore the shape support)
            min_width, max_width = width_s
            min_length, max_length = length_s
            min_height, max_height = height_s

            if None in [
                min_width,
                max_width,
                min_length,
                max_length,
                min_height,
                max_height,
            ]:
                # Can't get a bound on one or more dimensions, abort
                return None, None

            min_bounds = np.array([min_width, min_length, min_height])
            max_bounds = np.array([max_width, max_length, max_height])

            # Extract a list of possible shapes
            if isinstance(self.shape, Shape):
                shapes = [self.shape]
            elif isinstance(self.shape, MultiplexerDistribution) and all(
                isinstance(opt, Shape) for opt in self.shape.options
            ):
                shapes = self.shape.options
            else:
                # Something we don't recognize, abort
                return None, None

            # Get the inradius for each shape with the min and max bounds
            min_distances = [
                MeshVolumeRegion(mesh=shape.mesh, dimensions=min_bounds).inradius
                for shape in shapes
            ]
            max_distances = [
                MeshVolumeRegion(mesh=shape.mesh, dimensions=max_bounds).inradius
                for shape in shapes
            ]

            distance_range = (min(min_distances), max(max_distances))

            return distance_range

        # Define a helper function that computes the actual inradius
        @distributionFunction(support=inradiusSupport)
        def inradiusActual(width, length, height, shape):
            return MeshVolumeRegion(
                mesh=shape.mesh, dimensions=(width, length, height)
            ).inradius

        # Return the inradius (possibly a distribution) with proper support information
        return inradiusActual(self.width, self.length, self.height, self.shape)

    @cached_property
    def planarInradius(self):
        """A lower bound on the planar inradius of this object.

        This is defined as the inradius of the polygon of the occupiedSpace
        of this object projected into the XY plane, assuming that pitch and
        roll are both 0.
        """

        # Define a helper function that computes the support of the inradius,
        # given the sub supports.
        def planarInradiusSupport(width_s, length_s, shape_s):
            # Unpack the dimension supports (and ignore the shape support)
            min_width, max_width = width_s
            min_length, max_length = length_s

            if None in [min_width, max_width, min_length, max_length]:
                # Can't get a bound on one or more dimensions, abort
                return None, None

            min_bounds = np.array([min_width, min_length, 1])
            max_bounds = np.array([max_width, max_length, 1])

            # Extract a list of possible shapes
            if isinstance(self.shape, Shape):
                shapes = [self.shape]
            elif isinstance(self.shape, MultiplexerDistribution) and all(
                isinstance(opt, Shape) for opt in self.shape.options
            ):
                shapes = self.shape.options
            else:
                # Something we don't recognize, abort
                return None, None

            # Get the inradius of the projected for each shape with the min and max bounds
            min_distances = [
                MeshVolumeRegion(
                    mesh=shape.mesh, dimensions=min_bounds
                ).boundingPolygon.inradius
                for shape in shapes
            ]
            max_distances = [
                MeshVolumeRegion(
                    mesh=shape.mesh, dimensions=max_bounds
                ).boundingPolygon.inradius
                for shape in shapes
            ]

            distance_range = (min(min_distances), max(max_distances))

            return distance_range

        # Define a helper function that computes the actual planarInradius
        @distributionFunction(support=planarInradiusSupport)
        def planarInradiusActual(width, length, shape):
            return MeshVolumeRegion(
                mesh=shape.mesh, dimensions=(width, length, 1)
            ).boundingPolygon.inradius

        # Return the planar inradius (possibly a distribution) with proper support information
        return planarInradiusActual(self.width, self.length, self.shape)

    @cached_property
    def surface(self):
        """A region containing the entire surface of this object"""
        return self.occupiedSpace.getSurfaceRegion()

    @cached_property
    def onSurface(self):
        """The surface used by the ``on`` specifier.

        This region is used to sample position when
        another object is placed ``on`` this object. By default
        the top surface of this object (`topSurface`), but can
        be overwritten by subclasses.
        """
        return self.topSurface

    @cached_property
    def topSurface(self):
        """A region containing the top surface of this object

        For how this surface is computed, see `defaultSideSurface`.
        """
        return defaultSideSurface(
            self.occupiedSpace,
            dimension=2,
            positive=True,
            thresholds=self.sideComponentThresholds,
        )

    @cached_property
    def rightSurface(self):
        """A region containing the right surface of this object

        For how this surface is computed, see `defaultSideSurface`.
        """
        return defaultSideSurface(
            self.occupiedSpace,
            dimension=0,
            positive=True,
            thresholds=self.sideComponentThresholds,
        )

    @cached_property
    def leftSurface(self):
        """A region containing the left surface of this object

        For how this surface is computed, see `defaultSideSurface`.
        """
        return defaultSideSurface(
            self.occupiedSpace,
            dimension=0,
            positive=False,
            thresholds=self.sideComponentThresholds,
        )

    @cached_property
    def frontSurface(self):
        """A region containing the front surface of this object

        For how this surface is computed, see `defaultSideSurface`.
        """
        return defaultSideSurface(
            self.occupiedSpace,
            dimension=1,
            positive=True,
            thresholds=self.sideComponentThresholds,
        )

    @cached_property
    def backSurface(self):
        """A region containing the back surface of this object

        For how this surface is computed, see `defaultSideSurface`.
        """
        return defaultSideSurface(
            self.occupiedSpace,
            dimension=1,
            positive=False,
            thresholds=self.sideComponentThresholds,
        )

    @cached_property
    def bottomSurface(self):
        """A region containing the bottom surface of this object

        For how this surface is computed, see `defaultSideSurface`.
        """
        return defaultSideSurface(
            self.occupiedSpace,
            dimension=2,
            positive=False,
            thresholds=self.sideComponentThresholds,
        )

    def show3D(self, viewer, highlight=False):
        if needsSampling(self):
            raise RuntimeError("tried to show() symbolic Object")

        # Render the object
        object_mesh = self.occupiedSpace.mesh.copy()

        if highlight:
            object_mesh.visual.face_colors = [30, 179, 0, 255]
        elif self.color is not None:
            if len(self.color) == 3:
                r, g, b = self.color
                a = 1
            elif len(self.color) == 4:
                r, g, b, a = self.color
            else:
                assert False

            object_mesh.visual.face_colors = [255 * r, 255 * g, 255 * b, 255 * a]

        viewer.add_geometry(object_mesh)

        if self.showVisibleRegion:
            view_region_mesh = self.visibleRegion.mesh

            edges = view_region_mesh.face_adjacency_edges[
                view_region_mesh.face_adjacency_angles > np.radians(0.1)
            ]
            vertices = view_region_mesh.vertices

            edge_path = trimesh.path.Path3D(
                **trimesh.path.exchange.misc.edges_to_path(edges, vertices)
            )

            edge_path.colors = [
                [30, 30, 150, 255] for _ in range(len(edge_path.entities))
            ]

            viewer.add_geometry(edge_path)

    def show2D(self, workspace, plt, highlight=False):
        if needsSampling(self):
            raise RuntimeError("tried to show() symbolic Object")
        pos = self.position
        spos = workspace.scenicToSchematicCoords(pos)

        if highlight:
            # Circle around object
            rad = 1.5 * max(self.width, self.length)
            c = plt.Circle(spos, rad, color="g", fill=False)
            plt.gca().add_artist(c)
            # View cone
            ha = self.viewAngle / 2.0
            camera = self.position.offsetRotated(self.heading, self.cameraOffset)
            cpos = workspace.scenicToSchematicCoords(camera)
            for angle in (-ha, ha):
                p = camera.offsetRadially(20, self.heading + angle)
                edge = [cpos, workspace.scenicToSchematicCoords(p)]
                x, y = zip(*edge)
                plt.plot(x, y, "b:")

        corners = [
            workspace.scenicToSchematicCoords(corner) for corner in self._corners2D
        ]
        x, y = zip(*corners)
        color = self.color if hasattr(self, "color") else (1, 0, 0)
        plt.fill(x, y, color=color)

        frontMid = averageVectors(corners[0], corners[1])
        baseTriangle = [frontMid, corners[2], corners[3]]
        triangle = [averageVectors(p, spos, weight=0.5) for p in baseTriangle]
        x, y = zip(*triangle)
        plt.fill(x, y, "w")
        plt.plot(x + (x[0],), y + (y[0],), color="k", linewidth=1)

    @cached_property
    def _isPlanarBox(self):
        """Whether this object is a box aligned with the XY plane."""
        return (
            isinstance(self.shape, BoxShape)
            and self.orientation.pitch == 0
            and self.orientation.roll == 0
        )

    @cached_property
    def _boundingPolygon(self):
        # Fast case for planar boxes
        if self._isPlanarBox:
            width, length = self.width, self.length
            pos = self.position
            yaw = self.orientation.yaw
            cyaw, syaw = math.cos(yaw), math.sin(yaw)
            matrix = [
                width * cyaw,
                -length * syaw,
                width * syaw,
                length * cyaw,
                pos[0],
                pos[1],
            ]
            return shapely.affinity.affine_transform(_unitBox, matrix)

        return self.occupiedSpace._boundingPolygon


_unitBox = shapely.geometry.Polygon(((0.5, 0.5), (-0.5, 0.5), (-0.5, -0.5), (0.5, -0.5)))


@distributionFunction
def defaultSideSurface(
    occupiedSpace, dimension, positive, thresholds
) -> MeshSurfaceRegion:
    """Extracts a side surface from the occupiedSpace of an object.

    This function is the default implementation for computing a region
    representing a side surface of an object. This is done by keeping only the
    faces of the object's ``occupiedSpace`` mesh that have normal
    vectors with a large/small enough x,y, or z component. For example,
    for the front surface of an object we would would keep all faces that
    had a normal vector with y component greater than ``thresholds[1][1]``
    and for the back surface of an object we would keep all faces that
    had a normal vector with y component less than ``thresholds[1][0]``.

    Args:
        occupiedSpace: The `occupiedSpace` region of the object to
          extract the side surface from.
        dimension: The target dimension who's component will be checked.
        positive: If `False`, the target component must be less than
          the first value in the appropriate tuple. If `True`, the
          component must be greater than the second value in the
          appropriate tuple.
        thresholds: A 3-tuple of 2-tuples, one for each dimension (x,y,z),
          with each tuple containing the thresholds for a non-positive and
          positive side, respectively, in each dimension.
        on_dimension: The on_dimension to be passed to the created surface.
    """
    # Extract mesh from object
    obj_mesh = occupiedSpace.mesh.copy()

    # Extract appropriate thresholds
    threshold = thresholds[dimension][int(positive)]

    # Drop all faces whose normal vector do not have a sufficiently
    # large component.
    face_normal_vals = obj_mesh.face_normals[:, dimension]
    if positive:
        face_mask = face_normal_vals >= threshold
    else:
        face_mask = face_normal_vals <= threshold

    obj_mesh.faces = obj_mesh.faces[face_mask]
    obj_mesh.remove_unreferenced_vertices()

    # Check if the resulting surface is empty and return an appropriate region.
    if not obj_mesh.is_empty:
        return MeshSurfaceRegion(mesh=obj_mesh, centerMesh=False)
    else:
        return EmptyRegion(name="EmptyTopSurface")


def enableDynamicProxyFor(obj):
    object.__setattr__(obj, "_dynamicProxy", obj._copyWith())


def setDynamicProxyFor(obj, proxy):
    object.__setattr__(obj, "_dynamicProxy", proxy)


def disableDynamicProxyFor(obj):
    object.__setattr__(obj, "_dynamicProxy", obj)


## 2D Compatibility Classes


class Point2D(Point):
    """A 2D version of `Point`, used for backwards compatibility with Scenic 2.0"""

    _scenic_properties = {}
    _3DClass = Point

    @cached_property
    def visibleRegion(self):
        """The :term:`visible region` of this 2D point.

        The visible region of a `Point` is a disc centered at its ``position`` with
        radius ``visibleDistance``.
        """
        return CircularRegion(self.position, self.visibleDistance)

    def _canSee2D(self, other):
        if isinstance(other, Object2D):
            return self.visibleRegion.polygons.intersects(other._boundingPolygon)
        elif isinstance(other, (Vector, Point2D)):
            return self.visibleRegion.containsPoint(toVector(other))
        else:
            assert False, other

    def canSee(self, other, occludingObjects=tuple()):
        # Fast path when there is no occlusion (default in 2D mode).
        if not occludingObjects:
            return self._canSee2D(other)

        # With occlusion, fall back to the general case.
        return self._3DClass.canSee(self, other, occludingObjects)


class OrientedPoint2D(Point2D, OrientedPoint):
    """A 2D version of `OrientedPoint`, used for backwards compatibility with Scenic 2.0"""

    _scenic_properties = {}
    _3DClass = OrientedPoint

    def __init_subclass__(cls):
        if cls.__dict__.get("_props_transformed", False):
            # Can get here when cls is unpickled (the transformed version was pickled)
            pass
        else:
            # Mark class as being transformed.
            # To work around https://github.com/uqfoundation/dill/issues/612,
            # use a different truthy value for each class.
            cls._props_transformed = str(cls)

            props = cls._scenic_properties
            # Raise error if parentOrientation and heading already defined
            if "parentOrientation" in props and "heading" in props:
                raise RuntimeError(
                    f'{cls.__name__} defines both "parentOrientation" and "heading"'
                )

            # Map certain properties to their 3D analog
            if "heading" in props:
                props["parentOrientation"] = props["heading"]
                del props["heading"]

        super().__init_subclass__()

    @classmethod
    def _prepareSpecifiers(cls, specifiers):
        # Map certain specifiers to their 3D analog
        newspecs = []
        for spec in specifiers:
            # Map "with heading x" to "facing x"
            if spec.name == "With(heading)" and tuple(spec.priorities) == ("heading",):
                import scenic.syntax.veneer as veneer

                newspecs.append(veneer.Facing(spec.value["heading"]))
            else:
                newspecs.append(spec)
        return newspecs

    @cached_property
    def visibleRegion(self):
        """The :term:`visible region` of this 2D oriented point.

        The visible region of an `OrientedPoint` is a sector of the disc centered at its
        ``position`` with radius ``visibleDistance``, oriented along ``heading`` and
        subtending an angle of ``viewAngle``.
        """
        return SectorRegion(
            self.position, self.visibleDistance, self.heading, self.viewAngle
        )


class Object2D(OrientedPoint2D, Object):
    """A 2D version of `Object`, used for backwards compatibility with Scenic 2.0"""

    _scenic_properties = {
        "baseOffset": (0, 0, 0),
        "contactTolerance": 0,
        "requireVisible": True,
        "occluding": False,
        "height": PropertyDefault(
            ("width", "length"), {}, lambda self: max(self.width, self.length)
        ),
    }
    _3DClass = Object

    @classmethod
    def _specify(cls, context, prop, value):
        # If position is being set, set z value to 0
        if prop == "position":
            value = toVector(value, f'"{prop}" of {cls.__name__} not a vector')
            if needsSampling(value.z) or value.z != 0:
                # only modify value if necessary, to keep expression forest simpler
                value = toVector((value.x, value.y, 0))

        if prop == "shape" and not isinstance(value, BoxShape):
            raise InvalidScenarioError(
                "non-box shapes not allowed in 2D compatibility mode"
            )

        super()._specify(context, prop, value)

    @cached_property
    def visibleRegion(self):
        """The :term:`visible region` of this 2D object.

        The visible region of a 2D `Object` is a circular sector as for `OrientedPoint`,
        except that the base of the sector may be offset from ``position`` by the
        ``cameraOffset`` property (to allow modeling cameras which are not located at the
        center of the object).
        """
        camera = self.position.offsetRotated(self.heading, self.cameraOffset)
        return SectorRegion(camera, self.visibleDistance, self.heading, self.viewAngle)



"""Objects representing propositions that can be used to specify conditions"""

from functools import reduce
import operator
from typing import List

import rv_ltl

from scenic.core.errors import InvalidScenarioError
from scenic.core.lazy_eval import needsLazyEvaluation


class PropositionMonitor:
    def __init__(self, proposition: "PropositionNode") -> None:
        self._proposition = proposition
        self._monitor = proposition.ltl_node.create_monitor()

    def update(self):
        atomic_propositions = self._proposition.atomics()
        state = {}
        for ap in atomic_propositions:
            b = ap.closure()
            if needsLazyEvaluation(b):
                raise InvalidScenarioError(
                    f"value undefined outside of object definition"
                )
            state[str(ap.syntax_id)] = b
        self._monitor.update(state)
        return self._monitor.evaluate()


class PropositionNode:
    """Base class for temporal and non-temporal propositions"""

    def __init__(self, ltl_node) -> None:
        self.ltl_node = ltl_node

        self.is_temporal = False
        """tells if the proposition is temporal"""

    def check_constrains_sampling(self):
        """Checks if the proposition can be used for pruning.

        A requirement can be used for pruning if it is evaluated on the scene generation phase before simulation, and
        violation in that phase immediately results in discarding the scene and regenerating a new one.
        For simplicity, we currently check two special cases:
        1. requirements with no temporal requirement
        2. requirements with only one always operator on top-level

        Returns:
            bool: True if the requirement is one of the forms above. False otherwise.
        """
        node = self

        # if always is on top-level, check what's inside
        if isinstance(node, Always):
            node = node.req

        # eligible if none of the nodes was temporal
        eligible = all(not n.is_temporal for n in node.flatten())

        return eligible

    @property
    def children(self) -> List["PropositionNode"]:
        """Returns all children of proposition tree.

        Returns:
            list: proposition nodes that are directly under this node
        """
        return []

    def flatten(self) -> List["PropositionNode"]:
        """Flattens the tree and return the list of nodes.

        Returns:
            list: list of all children nodes
        """
        return [self] + reduce(
            operator.concat, [node.flatten() for node in self.children], []
        )

    def atomics(self) -> List["Atomic"]:
        return list(filter(lambda n: isinstance(n, Atomic), self.flatten()))

    def create_monitor(self) -> rv_ltl.Monitor:
        return PropositionMonitor(self)

    def evaluate(self):
        raise RuntimeError(
            "This proposition contains temporal operators and can only be evaluated using monitor"
        )

    @property
    def has_temporal_operator(self):
        node = self
        has_temporal_op = any(n.is_temporal for n in node.flatten())
        return has_temporal_op


class Atomic(PropositionNode):
    def __init__(self, closure, syntax_id):
        ap = rv_ltl.Atomic(identifier=str(syntax_id))
        super().__init__(ap)
        self.syntax_id = syntax_id
        self.closure = closure

    def __str__(self):
        return (
            f"(AP syntax_id={self.syntax_id})" if self.syntax_id is not None else "(AP)"
        )

    def evaluate(self):
        return self.closure()


class UnaryProposition(PropositionNode):
    """Base class for temporal unary operators"""

    @property
    def children(self):
        return [self.req]


class Always(UnaryProposition):
    def __init__(self, req: PropositionNode):
        ltl_node = rv_ltl.Always(req.ltl_node)
        super().__init__(ltl_node)
        self.req = req
        self.is_temporal = True

    def __str__(self):
        return f"(Always {str(self.req)})"


class Eventually(UnaryProposition):
    def __init__(self, req: PropositionNode):
        ltl_node = rv_ltl.Eventually(req.ltl_node)
        super().__init__(ltl_node)
        self.req = req
        self.is_temporal = True

    def __str__(self):
        return f"(Eventually {str(self.req)})"


class Next(UnaryProposition):
    def __init__(self, req: PropositionNode):
        ltl_node = rv_ltl.Next(req.ltl_node)
        super().__init__(ltl_node)
        self.req = req
        self.is_temporal = True

    def __str__(self):
        return f"(Next {str(self.req)})"


class Not(UnaryProposition):
    def __init__(self, req: PropositionNode):
        ltl_node = rv_ltl.Not(req.ltl_node)
        super().__init__(ltl_node)
        self.req = req

    def __str__(self):
        return f"(Not {str(self.req)})"

    def evaluate(self):
        return not self.req.evaluate()


class And(PropositionNode):
    def __init__(self, reqs):
        ltl_node = rv_ltl.And(*[req.ltl_node for req in reqs])
        super().__init__(ltl_node)
        self.reqs = reqs

    def __str__(self):
        return " and ".join([f"{str(req)}" for req in self.reqs])

    @property
    def children(self):
        return self.reqs

    def evaluate(self):
        return reduce(operator.and_, [node.evaluate() for node in self.reqs], True)


class Or(PropositionNode):
    def __init__(self, reqs):
        ltl_node = rv_ltl.Or(*[req.ltl_node for req in reqs])
        super().__init__(ltl_node)
        self.reqs = reqs

    def __str__(self):
        return " or ".join([f"{str(req)}" for req in self.reqs])

    @property
    def children(self):
        return self.reqs

    def evaluate(self):
        return reduce(operator.or_, [node.evaluate() for node in self.reqs], False)


class Until(PropositionNode):
    def __init__(self, lhs, rhs) -> None:
        self.lhs = lhs
        self.rhs = rhs
        ltl_node = rv_ltl.Until(lhs.ltl_node, rhs.ltl_node)
        super().__init__(ltl_node)
        self.is_temporal = True

    def __str__(self):
        return f"({self.lhs} until {self.rhs})"

    @property
    def children(self):
        return [self.lhs, self.rhs]


class Implies(PropositionNode):
    def __init__(self, lhs, rhs) -> None:
        self.lhs = lhs
        self.rhs = rhs
        ltl_node = rv_ltl.Implies(lhs.ltl_node, rhs.ltl_node)
        super().__init__(ltl_node)

    def __str__(self):
        return f"({self.lhs} implies {self.rhs})"

    @property
    def children(self):
        return [self.lhs, self.rhs]



"""General code for invocables, i.e. behaviors, monitors, and modular scenarios."""

import enum
import types

from scenic.core.distributions import Options
from scenic.core.errors import InvalidScenarioError
from scenic.core.lazy_eval import DelayedArgument, needsLazyEvaluation

from .guards import GuardViolation
from .utils import RejectSimulationException


class Invocable:
    """Abstract class with common code for behaviors and modular scenarios.

    Both of these types of objects can be called like functions, can have guards, and can
    suspend their own execution to invoke sub-behaviors/scenarios.
    """

    def __init__(self, *args, **kwargs):
        import scenic.syntax.veneer as veneer

        if veneer.evaluatingGuard:
            raise InvalidScenarioError(
                "tried to invoke behavior/scenario from inside guard or interrupt condition"
            )
        self._args = args
        self._kwargs = kwargs
        self._agent = None
        self._runningIterator = None
        self._isRunning = False

    def _start(self):
        assert not self._isRunning
        self._isRunning = True
        self._finalizeArguments()

    def _step(self):
        assert self._isRunning

    def _stop(self, reason=None):
        assert self._isRunning
        self._isRunning = False

    def _finalizeArguments(self):
        # Evaluate any lazy arguments whose evaluation was deferred until just before
        # this invocable starts running.
        args = []
        for arg in self._args:
            if needsLazyEvaluation(arg):
                assert isinstance(arg, DelayedArgument)
                args.append(arg.evaluateInner(None))
            else:
                args.append(arg)
        self._args = tuple(args)
        for name, arg in self._kwargs.items():
            if needsLazyEvaluation(arg):
                assert isinstance(arg, DelayedArgument)
                self._kwargs[name] = arg.evaluateInner(None)

    def _invokeSubBehavior(self, agent, subs, modifier=None, schedule=None):
        def pickEnabledInvocable(opts):
            enabled = {}
            if isinstance(opts, dict):
                for sub, weight in opts.items():
                    if sub._isEnabledForAgent(agent):
                        enabled[sub] = weight
            else:
                for sub in opts:
                    if sub._isEnabledForAgent(agent):
                        enabled[sub] = 1
            if not enabled:
                raise RejectSimulationException('deadlock in "do choose/shuffle"')
            if len(enabled) == 1:
                choice = list(enabled)[0]
            else:
                choice = Options(enabled)
            return choice

        scheduler = None
        if schedule == "choose":
            if len(subs) == 1 and isinstance(subs[0], dict):
                subs = subs[0]
            subs = (pickEnabledInvocable(subs),)
        elif schedule == "shuffle":
            if len(subs) == 1 and isinstance(subs[0], dict):
                subs = subs[0]
            else:
                subs = {item: 1 for item in subs}

            def scheduler():
                while subs:
                    choice = pickEnabledInvocable(subs)
                    subs.pop(choice)
                    yield from self._invokeInner(agent, (choice,))

        else:
            assert schedule is None
        if not scheduler:

            def scheduler():
                yield from self._invokeInner(agent, subs)

        if modifier:
            if modifier.name == "for":  # do X for Y [seconds | steps]
                import scenic.syntax.veneer as veneer

                timeLimit = modifier.value
                if not isinstance(timeLimit, (float, int)):
                    raise TypeError('"do X for Y" with Y not a number')
                assert modifier.terminator in (None, "seconds", "steps")
                if modifier.terminator != "steps":
                    timeLimit /= veneer.currentSimulation.timestep
                startTime = veneer.currentSimulation.currentTime
                condition = (
                    lambda: veneer.currentSimulation.currentTime - startTime >= timeLimit
                )
            elif modifier.name == "until":  # do X until Y
                condition = modifier.value
            else:
                raise RuntimeError(
                    f"internal parsing error: impossible modifier {modifier}"
                )

            def body(behavior, agent):
                yield from scheduler()

            def handler(behavior, agent):
                for sub in subs:
                    if sub._isRunning:
                        sub._stop(f'"{modifier.name}" condition met')
                return BlockConclusion.ABORT

            yield from runTryInterrupt(self, agent, body, [condition], [handler])
        else:
            yield from scheduler()

    def _invokeInner(self, agent, subs):
        """Run the given sub-behavior/scenario(s) in parallel.

        Implemented by subclasses.
        """
        raise NotImplementedError

    def _checkAllPreconditions(self):
        self.checkPreconditions(self._agent, *self._args, **self._kwargs)
        self.checkInvariants(self._agent, *self._args, **self._kwargs)

    def _isEnabledForAgent(self, agent):
        assert not self._isRunning
        assert self._agent is None
        try:
            self._agent = agent  # in case `self` is used in a precondition
            self._checkAllPreconditions()
            return True
        except GuardViolation:
            return False
        finally:
            self._agent = None


# Try-interrupt blocks


def runTryInterrupt(behavior, agent, body, conditions, handlers):
    body = InterruptBlock(None, body)
    interrupts = [InterruptBlock(c, h) for c, h in zip(conditions, handlers)]
    while True:
        # find next block to run, if any
        block = body
        for interrupt in interrupts:
            if interrupt.isEnabled or interrupt.isRunning:
                block = interrupt
                break
        result, concluded = block.step(behavior, agent)
        if concluded:
            if result is BlockConclusion.FINISHED and block is not body:
                continue  # interrupt handler finished
            else:
                return result  # entire try-interrupt statement will terminate
        else:
            yield result
            behavior.checkInvariants(None, *behavior._args, **behavior._kwargs)


@enum.unique
class BlockConclusion(enum.Enum):
    FINISHED = enum.auto()
    ABORT = enum.auto()
    RETURN = enum.auto()
    BREAK = enum.auto()
    CONTINUE = enum.auto()

    def __call__(self, value):
        self.return_value = value
        return self


class InterruptBlock:
    def __init__(self, condition, body):
        self.condition = condition
        self.body = body
        self.runningIterator = None

    @property
    def isEnabled(self):
        import scenic.syntax.veneer as veneer

        with veneer.executeInGuard():
            result = self.condition()
            if isinstance(result, DelayedArgument):
                # Condition cannot yet be evaluated because it depends on a scenario
                # local not yet initialized; we consider it to be false.
                return False
            return bool(result)

    @property
    def isRunning(self):
        return self.runningIterator is not None

    def step(self, behavior, agent):
        if not self.runningIterator:
            it = self.body(behavior, agent)
            if not isinstance(it, types.GeneratorType):
                return (it, True)
            self.runningIterator = it
        try:
            result = self.runningIterator.send(None)
            return (result, False)
        except StopIteration as e:
            self.runningIterator = None
            return (e.value, True)



"""Various utilities for working with Webots scenarios."""

import abc
import math
from typing import List

import numpy as np
from scipy.spatial.transform import Rotation as R

from scenic.core.geometry import normalizeAngle
from scenic.core.vectors import Orientation

## Coordinate system transformations


class WebotsCoordinateSystem:
    """A Webots coordinate system into which Scenic positions can be converted.

    See the Webots documentation of `WorldInfo.coordinateSystem`_ for a discussion
    of the possible coordinate systems. Since Webots R2022a, the default coordinate
    axis convention is ENU (X-Y-Z=East-North-Up), which is the same as Scenic's.

    .. _WorldInfo.coordinateSystem: https://cyberbotics.com/doc/reference/worldinfo
    """

    def __init__(self, system="ENU"):
        self.system = system
        axisMap = (system.find("E"), system.find("N"), system.find("U"))
        if len(system) != 3 or -1 in axisMap:
            raise RuntimeError("unknown kind of Webots coordinate system")
        self.axisMap = axisMap
        self.invAxisMap = (axisMap.index(0), axisMap.index(1), axisMap.index(2))
        self.upAxis = [0, 0, 0]
        self.upAxis[self.axisMap[2]] = 1
        self.leftHanded = system[:2] in ("EU", "NE", "UN")
        if self.leftHanded:
            self.mult = [1, -1, 1]
            self.invMult = [1, 1, -1]
        else:
            self.mult = self.invMult = [1, 1, 1]

    def positionToScenic(self, pos):
        """Convert a Webots position to a Scenic position."""
        return list(self.mult[i] * pos[self.axisMap[i]] for i in range(3))

    def positionFromScenic(self, pos):
        """Convert a Scenic position to a Webots position."""
        return list(self.invMult[i] * pos[self.invAxisMap[i]] for i in range(3))

    def orientationFromScenic(
        self, orientation: Orientation, offset: Orientation
    ) -> List[float]:
        # TODO(shun): Support other coordinate systems
        if self.system != "ENU":
            raise ValueError("Coordinate systems other than ENU is not fully supported")

        target = orientation * offset
        r = target.getRotation()
        rotvec = r.as_rotvec()
        if rotvec.tolist() == [0, 0, 0]:
            webotsRotation = [1.0, 0.0, 0.0, 0.0]
        else:
            norm = np.linalg.norm(rotvec)
            rotvec = rotvec / norm
            webotsRotation = rotvec.tolist() + [norm]
        return webotsRotation

    def orientationToScenic(
        self, webotsOrientation: List[float], offset: Orientation
    ) -> Orientation:
        # TODO(shun): Support other coordinate systems
        if self.system != "ENU":
            raise ValueError("Coordinate systems other than ENU is not fully supported")

        # webotsOrientation is a list of length 4 whose first three values are the normalized rotation axis and
        # the fourth value is the rotation angle in radian
        angle = webotsOrientation[3]
        rotvec = np.array(webotsOrientation[0:3]) * angle
        target = Orientation(R.from_rotvec(rotvec.tolist()))
        orientation = target * offset.inverse
        return orientation


ENU = WebotsCoordinateSystem("ENU")  #: The ENU coordinate system (the Webots default).
NUE = WebotsCoordinateSystem("NUE")  #: The NUE coordinate system.
EUN = WebotsCoordinateSystem("EUN")  #: The EUN coordinate system.



"""Behaviors and monitors."""

import functools
import inspect
import itertools
import sys
import warnings

from scenic.core.distributions import Samplable, toDistribution
import scenic.core.dynamics as dynamics
from scenic.core.errors import InvalidScenarioError
from scenic.core.type_support import CoercionFailure
from scenic.core.utils import alarm

from .invocables import Invocable
from .utils import StuckBehaviorWarning


class Behavior(Invocable, Samplable):
    """Dynamic behaviors of agents.

    Behavior statements are translated into definitions of subclasses of this class.
    """

    _noActionsMsg = (
        'does not take any actions (perhaps you forgot to use "take" or "do"?)'
    )

    def __init_subclass__(cls):
        if "__signature__" in cls.__dict__:
            # We're unpickling a behavior; skip this step.
            return

        if cls.__module__ is not __name__:
            import scenic.syntax.veneer as veneer

            if veneer.currentScenario:
                veneer.currentScenario._behaviors.append(cls)

            target = cls.makeGenerator
            target = functools.partial(target, 0, 0)  # account for Scenic-inserted args
            cls.__signature__ = inspect.signature(target)

    def __init__(self, *args, **kwargs):
        args = tuple(toDistribution(arg) for arg in args)
        kwargs = {name: toDistribution(arg) for name, arg in kwargs.items()}

        # Validate arguments to the behavior
        sig = inspect.signature(self.makeGenerator)
        sig.bind(None, *args, **kwargs)  # raises TypeError on incompatible arguments
        Samplable.__init__(self, itertools.chain(args, kwargs.values()))
        Invocable.__init__(self, *args, **kwargs)

        if not inspect.isgeneratorfunction(self.makeGenerator):
            raise InvalidScenarioError(f"{self} {self._noActionsMsg}")

    @classmethod
    def _canCoerceType(cls, ty):
        return issubclass(ty, cls) or ty in (type, type(None))

    @classmethod
    def _coerce(cls, thing):
        if thing is None or isinstance(thing, cls):
            return thing
        elif issubclass(thing, cls):
            return thing()
        else:
            raise CoercionFailure(f"expected type of behavior, got {thing}")

    def sampleGiven(self, value):
        args = (value[arg] for arg in self._args)
        kwargs = {name: value[val] for name, val in self._kwargs.items()}
        return type(self)(*args, **kwargs)

    def _assignTo(self, agent):
        if self._agent and agent is self._agent._dynamicProxy:
            # Assigned again (e.g. by override) to same agent; do nothing.
            return
        if self._isRunning:
            raise InvalidScenarioError(
                f"tried to reuse behavior object {self} already assigned to {self._agent}"
            )
        self._start(agent)

    def _start(self, agent):
        super()._start()
        self._agent = agent
        self._runningIterator = self.makeGenerator(agent, *self._args, **self._kwargs)
        self._checkAllPreconditions()

    def _step(self):
        import scenic.syntax.veneer as veneer

        super()._step()
        assert self._runningIterator

        def alarmHandler(signum, frame):
            # NOTE: if using pytest-cov, sys.gettrace() set to CTracer(), but we still want timeout warnings enabled
            if sys.gettrace() and "coverage" not in str(type(sys.gettrace())):
                return  # skip the warning if we're in the debugger
            warnings.warn(
                f"the behavior {self} is taking a long time to take an action; "
                "maybe you have an infinite loop with no take/wait statements?",
                StuckBehaviorWarning,
            )

        timeout = dynamics.stuckBehaviorWarningTimeout
        with veneer.executeInBehavior(self), alarm(timeout, alarmHandler):
            try:
                actions = self._runningIterator.send(None)
            except StopIteration:
                actions = ()  # behavior ended early
        return actions

    def _stop(self, reason=None):
        super()._stop(reason)
        self._agent = None
        self._runningIterator = None

    @property
    def _isFinished(self):
        return self._runningIterator is None

    def _invokeInner(self, agent, subs):
        import scenic.syntax.veneer as veneer

        assert len(subs) == 1
        sub = subs[0]
        if not isinstance(sub, Behavior):
            raise TypeError(f"expected a behavior, got {sub}")
        sub._start(agent)
        with veneer.executeInBehavior(sub):
            try:
                yield from sub._runningIterator
            finally:
                if sub._isRunning:
                    sub._stop()

    def __repr__(self):
        items = itertools.chain(
            (repr(arg) for arg in self._args),
            (f"{key}={repr(val)}" for key, val in self._kwargs.items()),
        )
        allArgs = ", ".join(items)
        return f"{self.__class__.__name__}({allArgs})"


class Monitor(Behavior):
    """Monitors for dynamic simulations.

    Monitor statements are translated into definitions of subclasses of this class.
    """

    _noActionsMsg = 'does not take any actions (perhaps you forgot to use "wait"?)'

    def _start(self):
        return super()._start(None)



"""Actions for dynamic agents in the driving domain.

These actions are automatically imported when using the driving domain.

The `RegulatedControlAction` is based on code from the `CARLA`_ project, licensed under
the following terms:

    Copyright (c) 2018-2020 CVC.

    This work is licensed under the terms of the MIT license.
    For a copy, see <https://opensource.org/licenses/MIT>.

.. _CARLA: https://carla.org/
"""

import math

from scenic.core.simulators import Action
from scenic.core.vectors import Vector

## Mixin classes indicating support for various types of actions.


class Steers:
    """Mixin protocol for agents which can steer.

    Specifically, agents must support throttling, braking, steering, setting the hand
    brake, and going into reverse.
    """

    def setThrottle(self, throttle):
        raise NotImplementedError

    def setSteering(self, steering):
        raise NotImplementedError

    def setBraking(self, braking):
        raise NotImplementedError

    def setHandbrake(self, handbrake):
        raise NotImplementedError

    def setReverse(self, reverse):
        raise NotImplementedError


class Walks:
    """Mixin protocol for agents which can walk with a given direction and speed.

    We provide a simplistic implementation which directly sets the velocity of the agent.
    This implementation needs to be explicitly opted-into, since simulators may provide a
    more sophisticated API that properly animates pedestrians.
    """

    def setWalkingDirection(self, heading):
        velocity = Vector(0, self.speed).rotatedBy(heading)
        self.setVelocity(velocity)

    def setWalkingSpeed(self, speed):
        velocity = speed * self.velocity.normalized()
        self.setVelocity(velocity)


## Actions available to all agents


class SetPositionAction(Action):
    """Teleport an agent to the given position."""

    def __init__(self, pos: Vector):
        self.pos = pos

    def applyTo(self, obj, sim):
        obj.setPosition(self.pos, obj.elevation)


class OffsetAction(Action):
    """Teleports actor forward (in direction of its heading) by some offset."""

    def __init__(self, offset: Vector):
        self.offset = offset

    def applyTo(self, obj, sim):
        pos = obj.position.offsetRotated(obj.heading, self.offset)
        obj.setPosition(pos, obj.elevation)


class SetVelocityAction(Action):
    """Set the velocity of an agent."""

    def __init__(self, xVel: float, yVel: float, zVel: float = 0):
        self.velocity = (xVel, yVel, zVel)

    def applyTo(self, obj, sim):
        obj.setVelocity(self.velocity)


class SetSpeedAction(Action):
    """Set the speed of an agent (keeping its heading fixed)."""

    def __init__(self, speed: float):
        self.speed = speed

    def applyTo(self, obj, sim):
        vel = Vector(0, self.speed).rotatedBy(obj.heading)
        obj.setVelocity(vel)


## Actions available to vehicles which can steer


class SteeringAction(Action):
    """Abstract class for actions usable by agents which can steer.

    Such agents must implement the `Steers` protocol.
    """

    def canBeTakenBy(self, agent):
        return isinstance(agent, Steers)


class SetThrottleAction(SteeringAction):
    """Set the throttle.

    Arguments:
        throttle: Throttle value between 0 and 1.
    """

    def __init__(self, throttle: float):
        if not 0.0 <= throttle <= 1.0:
            raise RuntimeError("Throttle must be a float in range [0.0, 1.0].")
        self.throttle = throttle

    def applyTo(self, obj, sim):
        obj.setThrottle(self.throttle)


class SetSteerAction(SteeringAction):
    """Set the steering 'angle'.

    Arguments:
        steer: Steering 'angle' between -1 and 1.
    """

    def __init__(self, steer: float):
        if not -1.0 <= steer <= 1.0:
            raise RuntimeError("Steer must be a float in range [-1.0, 1.0].")
        self.steer = steer

    def applyTo(self, obj, sim):
        obj.setSteering(self.steer)


class SetBrakeAction(SteeringAction):
    """Set the amount of brake.

    Arguments:
        brake: Amount of braking between 0 and 1.
    """

    def __init__(self, brake: float):
        if not 0.0 <= brake <= 1.0:
            raise RuntimeError("Brake must be a float in range [0.0, 1.0].")
        self.brake = brake

    def applyTo(self, obj, sim):
        obj.setBraking(self.brake)


class SetHandBrakeAction(SteeringAction):
    """Set or release the hand brake.

    Arguments:
        handBrake: Whether or not the hand brake is set.
    """

    def __init__(self, handBrake: bool):
        if not isinstance(handBrake, bool):
            raise RuntimeError("Hand brake must be a boolean.")
        self.handbrake = handBrake

    def applyTo(self, obj, sim):
        obj.setHandbrake(self.handbrake)


class SetReverseAction(SteeringAction):
    """Engage or release reverse gear.

    Arguments:
        reverse: Whether or not the car is in reverse.
    """

    def __init__(self, reverse: bool):
        if not isinstance(reverse, bool):
            raise RuntimeError("Reverse must be a boolean.")
        self.reverse = reverse

    def applyTo(self, obj, sim):
        obj.setReverse(self.reverse)


class RegulatedControlAction(SteeringAction):
    """Regulated control of throttle, braking, and steering.

    Controls throttle and braking using one signal that may be positive or negative.
    Useful with simple controllers that output a single value.

    Arguments:
        throttle: Control signal for throttle and braking (will be clamped as below).
        steer: Control signal for steering (also clamped).
        past_steer: Previous steering signal, for regulating abrupt changes.
        max_throttle: Maximum value for **throttle**, when positive.
        max_brake: Maximum (absolute) value for **throttle**, when negative.
        max_steer: Maximum absolute value for **steer**.
    """

    def __init__(
        self,
        throttle: float,
        steer: float,
        past_steer: float,
        max_throttle: float = 0.5,
        max_brake: float = 0.5,
        max_steer: float = 0.8,
    ):
        if throttle > 0:
            throttle = min(throttle, max_throttle)
            brake = 0
        else:
            throttle = 0
            brake = min(abs(throttle), max_brake)

        # Steering regulation: changes cannot happen abruptly, can't steer too much.

        if steer > past_steer + 0.1:
            steer = past_steer + 0.1
        elif steer < past_steer - 0.1:
            steer = past_steer - 0.1

        if steer >= 0:
            steer = min(max_steer, steer)
        else:
            steer = max(-max_steer, steer)

        self.throttle, self.brake, self.steer = throttle, brake, steer

    def applyTo(self, obj, sim):
        obj.setThrottle(self.throttle)
        obj.setBraking(self.brake)
        obj.setSteering(self.steer)


## Actions available to agents that can walk


class WalkingAction(Action):
    """Abstract class for actions usable by agents which can walk.

    Such agents must implement the `Walks` protocol.
    """

    def canBeTakenBy(self, agent):
        return isinstance(agent, Walks)


class SetWalkingDirectionAction(WalkingAction):
    """Set the walking direction."""

    def __init__(self, heading):
        self.heading = heading

    def applyTo(self, obj, sim):
        obj.setWalkingDirection(self.heading)


class SetWalkingSpeedAction(WalkingAction):
    """Set the walking speed."""

    def __init__(self, speed):
        self.speed = speed

    def applyTo(self, obj, sim):
        obj.setWalkingSpeed(self.speed)



"""Scenic vectors and vector fields."""

from __future__ import annotations

import collections
import functools
import itertools
import math
from math import cos, sin
import numbers
import random
import struct
import typing
import warnings

import numpy
from scipy.spatial.transform import Rotation
import shapely.geometry

from scenic.core.distributions import (
    Distribution,
    MethodDistribution,
    RejectionException,
    Samplable,
    TupleDistribution,
    distributionFunction,
    distributionMethod,
    makeOperatorHandler,
    needsSampling,
)
from scenic.core.geometry import hypot, makeShapelyPoint, normalizeAngle
from scenic.core.lazy_eval import (
    isLazy,
    makeDelayedFunctionCall,
    needsLazyEvaluation,
    valueInContext,
)
from scenic.core.type_support import (
    CoercionFailure,
    canCoerceType,
    coerceToFloat,
    toOrientation,
)
from scenic.core.utils import argsToString, cached_property

# Suppress gimbal lock warning from scipy.spatial.transform.Rotation.
# N.B. due to the way the pytest works, the warning will still appear when
# running the test suite.
warnings.filterwarnings(
    "ignore",
    message="Gimbal lock detected. Setting third angle to zero",
    module="scenic.core.vectors",
)


class VectorDistribution(Distribution):
    """A distribution over Vectors."""

    _defaultValueType = None  # will be set after Vector is defined

    def toVector(self):
        return self


class VectorOperatorDistribution(VectorDistribution):
    """Vector version of OperatorDistribution."""

    def __init__(self, operator, obj, operands):
        super().__init__(obj, *operands)
        self.operator = operator
        self.object = obj
        self.operands = operands

    def sampleGiven(self, value):
        first = value[self.object]
        rest = (value[child] for child in self.operands)
        op = getattr(first, self.operator)
        return op(*rest)

    def evaluateInner(self, context):
        obj = valueInContext(self.object, context)
        operands = tuple(valueInContext(arg, context) for arg in self.operands)
        return VectorOperatorDistribution(self.operator, obj, operands)

    def __repr__(self):
        return f"{self.object!r}.{self.operator}({argsToString(self.operands)})"


class VectorMethodDistribution(VectorDistribution):
    """Vector version of MethodDistribution."""

    def __init__(self, method, obj, args, kwargs):
        super().__init__(*args, *kwargs.values())
        self.method = method
        self.object = obj
        self.arguments = args
        self.kwargs = kwargs

    def sampleGiven(self, value):
        args = (value[arg] for arg in self.arguments)
        kwargs = {name: value[arg] for name, arg in self.kwargs.items()}
        return self.method(self.object, *args, **kwargs)

    def evaluateInner(self, context):
        obj = valueInContext(self.object, context)
        arguments = tuple(valueInContext(arg, context) for arg in self.arguments)
        kwargs = {name: valueInContext(arg, context) for name, arg in self.kwargs.items()}
        return VectorMethodDistribution(self.method, obj, arguments, kwargs)

    def __repr__(self):
        args = argsToString(self.arguments, self.kwargs)
        return f"{self.object!r}.{self.method.__name__}({args})"


def scalarOperator(method):
    """Decorator for vector operators that yield scalars."""
    op = method.__name__
    setattr(VectorDistribution, op, makeOperatorHandler(op, float))

    @functools.wraps(method)
    def helper(self, *args, **kwargs):
        if any(needsSampling(arg) for arg in itertools.chain(args, kwargs.values())):
            return MethodDistribution(method, self, args, kwargs)
        else:
            return method(self, *args, **kwargs)

    return helper


def makeVectorOperatorHandler(op, zeroIdentity):
    def handler(self, *args):
        # If the zero vector is the identity for this operator, and we can tell
        # that the operand is zero, simplify the expression forest.
        if (
            zeroIdentity
            and not isLazy(args[0])
            and all(coord == 0 for coord in args[0].coordinates)
        ):
            return self

        # The general case.
        return VectorOperatorDistribution(op, self, args)

    return handler


def vectorOperator(method, preservesZero=False, zeroIdentity=False):
    """Decorator for vector operators that yield vectors."""
    op = method.__name__
    setattr(VectorDistribution, op, makeVectorOperatorHandler(op, zeroIdentity))

    @functools.wraps(method)
    def helper(self, *args):
        # If this operator preserves the zero vector, and we can tell that self
        # is zero, simplify the expression forest.
        if (
            preservesZero
            and not needsSampling(self)
            and all(coord == 0 for coord in self.coordinates)
        ):
            return self

        # General cases when the arguments are lazy.
        if any(needsSampling(arg) for arg in args):
            if needsSampling(self):
                return VectorOperatorDistribution(op, self, args)
            return VectorMethodDistribution(method, self, args, {})
        elif any(needsLazyEvaluation(arg) for arg in args):
            # see analogous comment in distributionFunction
            return makeDelayedFunctionCall(helper, args, {})

        # If zero is the identity, simplify when possible (see above).
        if (
            zeroIdentity
            and isinstance(args[0], (Vector, tuple, list, numpy.ndarray))
            and all(coord == 0 for coord in args[0])
        ):
            return self

        # General case when the arguments are not lazy.
        if needsSampling(self):
            return VectorOperatorDistribution(op, self, args)
        else:
            return method(self, *args)

    return helper


def zeroPreservingVectorOperator(method):
    return vectorOperator(method, preservesZero=True)


def zeroIdentityVectorOperator(method):
    return vectorOperator(method, zeroIdentity=True)


def vectorDistributionMethod(method):
    """Decorator for methods that produce vectors. See distributionMethod."""

    @functools.wraps(method)
    def helper(self, *args, **kwargs):
        if any(needsSampling(arg) for arg in itertools.chain(args, kwargs.values())):
            return VectorMethodDistribution(method, self, args, kwargs)
        elif any(
            needsLazyEvaluation(arg) for arg in itertools.chain(args, kwargs.values())
        ):
            # see analogous comment in distributionFunction
            return makeDelayedFunctionCall(helper, (self,) + args, kwargs)
        else:
            return method(self, *args, **kwargs)

    return helper


class Orientation:
    """An orientation in 3D space."""

    def __init__(self, rotation):
        if not isinstance(rotation, Rotation):
            raise TypeError(
                "Orientation's 'rotation' parameter must be a SciPy rotation."
                " Perhaps you want to use a factory method?"
            )
        self.r = rotation
        self.q = rotation.as_quat()

    @classmethod
    def fromQuaternion(cls, quaternion) -> Orientation:
        """Create an `Orientation` from a quaternion (of the form (x,y,z,w))"""
        r = Rotation.from_quat(quaternion)
        return cls(r)

    @classmethod
    @distributionFunction
    def fromEuler(cls, yaw, pitch, roll) -> Orientation:
        """Create an `Orientation` from yaw, pitch, and roll angles (in radians)."""
        return cls._fromEuler(yaw, pitch, roll)

    @classmethod
    def _fromEuler(cls, yaw, pitch, roll) -> Orientation:
        # Inner version of `fromEuler` which doesn't accept distributions.
        r = Rotation.from_euler("ZXY", [yaw, pitch, roll], degrees=False)
        return cls(r)

    @classmethod
    def _fromHeading(cls, heading) -> Orientation:
        # This method is faster than `from_euler` if we only have 1 angle.
        r = Rotation.from_rotvec([0, 0, heading], degrees=False)
        return cls(r)

    @property
    def w(self) -> float:
        return self.q[3]

    @property
    def x(self) -> float:
        return self.q[0]

    @property
    def y(self) -> float:
        return self.q[1]

    @property
    def z(self) -> float:
        return self.q[2]

    @property
    def yaw(self) -> float:
        """Yaw in the global coordinate system."""
        return self.eulerAngles[0]

    @property
    def pitch(self) -> float:
        """Pitch in the global coordinate system."""
        return self.eulerAngles[1]

    @property
    def roll(self) -> float:
        """Roll in the global coordinate system."""
        return self.eulerAngles[2]

    @staticmethod
    def _coerce(thing) -> Orientation:
        if isinstance(thing, (float, int)):  # fast path
            return Orientation._fromHeading(thing)
        elif isinstance(thing, Orientation):
            return thing
        elif hasattr(thing, "toOrientation"):
            return thing.toOrientation()
        elif isinstance(thing, Vector):
            return Orientation._fromEuler(*thing)
        elif isinstance(thing, (tuple, list)):
            if len(thing) != 3:
                raise CoercionFailure(
                    "Cannot coerce a tuple/list of length not 3 to an orientation"
                )
            return Orientation._fromEuler(*thing)
        elif canCoerceType(type(thing), float):
            return Orientation._fromHeading(coerceToFloat(thing))
        else:
            raise CoercionFailure

    @staticmethod
    def _canCoerceType(ty):
        if issubclass(ty, (tuple, list, Vector, Orientation)):
            return True
        return canCoerceType(ty, float) or hasattr(ty, "toOrientation")

    @cached_property
    def eulerAngles(self) -> typing.Tuple[float, float, float]:
        """Global intrinsic Euler angles yaw, pitch, roll."""
        return self.r.as_euler("ZXY", degrees=False)

    def _trimeshEulerAngles(self):
        return self.r.as_euler("xyz", degrees=False)

    def getRotation(self):
        return self.r

    @cached_property
    def inverse(self) -> Orientation:
        return Orientation(self.r.inv())

    @cached_property
    def _inverseRotation(self):
        return self.r.inv()

    # will be converted to a distributionMethod after the class definition
    def __mul__(self, other) -> Orientation:
        """Apply a rotation to this orientation, yielding a new orientation.

        As we represent orientations as intrinsic rotations, rotation A followed by rotation B is
        given by the quaternion product A*B, not B*A.

        See https://en.wikipedia.org/wiki/Davenport_chained_rotations#Conversion_to_extrinsic_rotations
        for more details.
        """
        if type(other) is not Orientation:
            return NotImplemented
        # Preserve existing orientation objects when possible to help pruning.
        if self == globalOrientation:
            return other
        if other == globalOrientation:
            return self
        return Orientation(self.r * other.r)

    @distributionMethod
    def __add__(self, other) -> Orientation:
        if isinstance(other, (float, int)):
            other = Orientation._fromHeading(other)
        elif type(other) is not Orientation:
            return NotImplemented
        return self * other

    @distributionMethod
    def __radd__(self, other) -> Orientation:
        if isinstance(other, (float, int)):
            other = Orientation._fromHeading(other)
        elif type(other) is not Orientation:
            return NotImplemented
        return other * self

    def __repr__(self):
        return f"Orientation.fromEuler{tuple(self.eulerAngles)!r}"

    def __hash__(self):
        return hash(tuple(self.q)) + hash(tuple(-self.q))

    @distributionFunction
    def localAnglesFor(self, orientation) -> typing.Tuple[float, float, float]:
        """Get local Euler angles for an orientation w.r.t. this orientation.

        That is, considering ``self`` as the parent orientation, find the Euler angles
        expressing the given orientation.
        """
        local = self.inverse * orientation
        return local.eulerAngles

    @distributionFunction
    def globalToLocalAngles(self, yaw, pitch, roll) -> typing.Tuple[float, float, float]:
        """Convert global Euler angles to local angles w.r.t. this orientation.

        Equivalent to `localAnglesFor` but takes Euler angles as input.
        """
        orientation = Orientation.fromEuler(yaw, pitch, roll)
        return self.localAnglesFor(orientation)

    def __eq__(self, other):
        if not isinstance(other, Orientation):
            return NotImplemented
        return numpy.array_equal(self.q, other.q) or numpy.array_equal(self.q, -other.q)

    def approxEq(self, other, tol=1e-10):
        if not isinstance(other, Orientation):
            return NotImplemented
        return abs(numpy.dot(self.q, other.q)) > 1 - tol

    @classmethod
    def encodeTo(cls, orientation, stream):
        stream.write(struct.pack("<dddd", *orientation.q))

    @classmethod
    def decodeFrom(cls, stream):
        # Quaternion constructor does not roundtrip so we manually
        # construct a rotation and pass that in.
        quaternion = struct.unpack("<dddd", stream.read(32))
        rotation = Rotation(quaternion, normalize=False)
        return cls(rotation)


globalOrientation = Orientation.fromEuler(0, 0, 0)

Orientation.__mul__ = distributionMethod(Orientation.__mul__, identity=globalOrientation)


def alwaysGlobalOrientation(orientation):
    """Whether this orientation is always aligned with the global coordinate system.

    Returns False if the orientation is a distribution or delayed argument, since
    then the value cannot be known at this time.
    """
    return isinstance(orientation, Orientation) and orientation == globalOrientation


class Vector(Samplable, collections.abc.Sequence):
    """A 3D vector, whose coordinates can be distributions."""

    def __init__(self, x, y, z=0):
        self.coordinates = (x, y, z)
        super().__init__(self.coordinates)

    @property
    def x(self) -> float:
        return self.coordinates[0]

    @property
    def y(self) -> float:
        return self.coordinates[1]

    @property
    def z(self) -> float:
        return self.coordinates[2]

    def toVector(self) -> Vector:
        return self

    @staticmethod
    def _canCoerceType(ty):
        return issubclass(ty, (tuple, list, numpy.ndarray)) or hasattr(ty, "toVector")

    @staticmethod
    def _coerce(thing) -> Vector:
        if isinstance(thing, (tuple, list, numpy.ndarray)):
            l = len(thing)
            if not 2 <= l <= 3:
                raise CoercionFailure(
                    "expected 2D/3D vector, got " f"{type(thing).__name__} of length {l}"
                )
            return Vector(*thing)
        else:
            return thing.toVector()

    def sampleGiven(self, value):
        return Vector(*(value[coord] for coord in self.coordinates))

    def evaluateInner(self, context):
        return Vector(*(valueInContext(coord, context) for coord in self.coordinates))

    @vectorOperator
    def applyRotation(self, rotation):
        if not isinstance(rotation, Orientation):
            return TypeError("rotation must be an Orientation")
        r = rotation.getRotation()
        return Vector(*r.apply(self.coordinates))

    @vectorOperator
    def sphericalCoordinates(self):
        """Returns this vector in spherical coordinates (rho, theta, phi)"""
        rho = math.hypot(self.x, self.y, self.z)
        theta = math.atan2(self.y, self.x) - math.pi / 2
        phi = math.atan2(self.z, math.hypot(self.x, self.y))
        return Vector(rho, theta, phi)

    @zeroPreservingVectorOperator
    def rotatedBy(self, angleOrOrientation) -> Vector:
        """Return a vector equal to this one rotated counterclockwise by angle/orientation."""
        if isinstance(angleOrOrientation, Orientation):
            return self.applyRotation(angleOrOrientation)
        x, y, z = self.x, self.y, self.z
        c, s = cos(angleOrOrientation), sin(angleOrOrientation)
        return Vector((c * x) - (s * y), (s * x) + (c * y), z)

    @vectorOperator
    def offsetRotated(self, angleOrOrientation, offset) -> Vector:
        ro = offset.rotatedBy(angleOrOrientation)
        return self + ro

    @vectorOperator
    def offsetLocally(self, orientation, offset) -> Vector:
        # Faster version of `offsetRotated` that only accepts Orientations.
        r = orientation.getRotation()
        ro = r.apply(offset)
        x, y, z = self
        ox, oy, oz = ro
        return Vector(x + ox, y + oy, z + oz)

    @vectorOperator
    def offsetRadially(self, radius, heading) -> Vector:
        return self.offsetRotated(heading, Vector(0, radius))

    @scalarOperator
    def distanceTo(self, other) -> float:
        if not isinstance(other, Vector):
            return other.distanceTo(self)
        dx, dy, dz = other.toVector() - self
        return math.hypot(dx, dy, dz)

    @scalarOperator
    def angleTo(self, other) -> float:
        return self.azimuthTo(other)

    @scalarOperator
    def azimuthTo(self, other) -> float:
        dx, dy, dz = other.toVector() - self
        return normalizeAngle(math.atan2(dy, dx) - (math.pi / 2))

    @scalarOperator
    def altitudeTo(self, other) -> float:
        dx, dy, dz = other.toVector() - self
        return normalizeAngle(math.atan2(dz, math.hypot(dx, dy)))

    @scalarOperator
    def angleWith(self, other) -> float:
        """Compute the signed angle between self and other.

        The angle is positive if other is counterclockwise of self (considering
        the smallest possible rotation to align them).
        """
        x, y = self.x, self.y
        ox, oy = other.x, other.y
        return normalizeAngle(math.atan2(oy, ox) - math.atan2(y, x))

    @scalarOperator
    def norm(self) -> float:
        return hypot(*self.coordinates)

    @scalarOperator
    def dot(self, other) -> float:
        x, y, z = self.x, self.y, self.z
        ox, oy, oz = other.x, other.y, other.z
        return (x * ox) + (y * oy) + (z * oz)

    @vectorOperator
    def cross(self, other) -> Vector:
        ax, ay, az = self.x, self.y, self.z
        bx, by, ba = other.x, other.y, other.z

        cx = ay * bz - az * by
        cy = az * bx - ax * bz
        cz = ax * by - ay * bx

        return Vector(cx, cy, cz)

    @vectorOperator
    def normalized(self) -> Vector:
        l = math.hypot(*self.coordinates)

        if l == 0:
            return Vector(0, 0, 0)

        return Vector(*(coord / l for coord in self.coordinates))

    @zeroIdentityVectorOperator
    def __add__(self, other) -> Vector:
        return Vector(self[0] + other[0], self[1] + other[1], self[2] + other[2])

    @zeroIdentityVectorOperator
    def __radd__(self, other) -> Vector:
        return Vector(self[0] + other[0], self[1] + other[1], self[2] + other[2])

    @zeroIdentityVectorOperator
    def __sub__(self, other) -> Vector:
        return Vector(self[0] - other[0], self[1] - other[1], self[2] - other[2])

    @vectorOperator
    def __rsub__(self, other) -> Vector:
        return Vector(other[0] - self[0], other[1] - self[1], other[2] - self[2])

    @vectorOperator
    def __mul__(self, other) -> Vector:
        return Vector(*(coord * other for coord in self.coordinates))

    def __rmul__(self, other) -> Vector:
        return self.__mul__(other)

    @vectorOperator
    def __truediv__(self, other) -> Vector:
        return Vector(*(coord / other for coord in self.coordinates))

    def __len__(self):
        return len(self.coordinates)

    def __getitem__(self, index):
        return self.coordinates[index]

    def __repr__(self):
        return f"Vector({self.x}, {self.y}, {self.z})"

    def __eq__(self, other):
        """Check Vector equality.

        A Vector is equal to another if their coordinates are equal,
        or if the other is a tuple/list that contains the coordinates of
        the Vector. For backwards compatibility a Vector is also equal
        to a tuple/list of length 2 that has a 0 z component if the Vector
        also has a 0 z component and the x and y components are equal.
        """
        if isinstance(other, Vector):
            return other.coordinates == self.coordinates
        elif isinstance(other, (tuple, list)):
            if len(other) == 2:
                return self.x == other[0] and self.y == other[1] and self.z == 0

            return tuple(other) == self.coordinates
        else:
            return NotImplemented

    def __hash__(self):
        return hash(self.coordinates)

    @classmethod
    def encodeTo(cls, vec, stream):
        stream.write(struct.pack("<ddd", *vec.coordinates))

    @classmethod
    def decodeFrom(cls, stream):
        return cls(*struct.unpack("<ddd", stream.read(24)))


VectorDistribution._defaultValueType = Vector


class OrientedVector(Vector):
    def __init__(self, x, y, z, heading):
        super().__init__(x, y, z)
        self.heading = heading

    @staticmethod
    @distributionFunction
    def make(position, heading) -> OrientedVector:
        return OrientedVector(*position, heading)

    def toHeading(self):
        return self.heading

    def evaluateInner(self, context):
        hdg = valueInContext(self.heading, context)
        return OrientedVector(
            *(valueInContext(coord, context) for coord in self.coordinates), hdg
        )

    def __eq__(self, other):
        if type(other) is not OrientedVector:
            return NotImplemented
        return other.coordinates == self.coordinates and other.heading == self.heading

    def __hash__(self):
        return hash((self.coordinates, self.heading))


class VectorField:
    """A vector field, providing an orientation at every point.

    Arguments:
        name (str): name for debugging.
        value: function computing the heading at the given `Vector`.
        minSteps (int): Minimum number of steps for `followFrom`; default 4.
        defaultStepSize (float): Default step size for `followFrom`; default 5.
            This is an upper bound: more steps will be taken as needed to ensure that no
            single step is longer than this value, but if the distance to travel is small
            then the steps may be smaller.
    """

    def __init__(self, name, value, minSteps=4, defaultStepSize=5):
        self.name = name
        self.value = value
        self.valueType = Orientation
        self.minSteps = minSteps
        self.defaultStepSize = defaultStepSize

    @distributionMethod
    def __getitem__(self, pos) -> Orientation:
        val = self.value(pos)
        if isinstance(val, numbers.Real):  # fast path
            return Orientation._fromHeading(val)
        if isLazy(val):
            raise ValueError(f"value function of {self.name} returned lazy value")
        return toOrientation(
            val, f"value function of {self.name} returned non-orientation"
        )

    @vectorDistributionMethod
    def followFrom(self, pos, dist, steps=None, stepSize=None):
        """Follow the field from a point for a given distance.

        Uses the forward Euler approximation, covering the given distance with
        equal-size steps. The number of steps can be given manually, or computed
        automatically from a desired step size.

        Arguments:
            pos (`Vector`): point to start from.
            dist (float): distance to travel.
            steps (int): number of steps to take, or :obj:`None` to compute the number of
                steps based on the distance (default :obj:`None`).
            stepSize (float): length used to compute how many steps to take, or
                :obj:`None` to use the field's default step size.
        """
        if steps is None:
            steps = self.minSteps
            stepSize = self.defaultStepSize if stepSize is None else stepSize
            if stepSize is not None:
                steps = max(steps, math.ceil(dist / stepSize))

        stepSize = dist / steps
        step = numpy.array([0, stepSize, 0])
        for i in range(steps):
            rot = self[pos].getRotation()
            pos += rot.apply(step)

        return Vector(*pos)

    @staticmethod
    def forUnionOf(regions, tolerance=0):
        """Creates a `PiecewiseVectorField` from the union of the given regions.

        If none of the regions have an orientation, returns :obj:`None` instead.
        """
        if any(reg.orientation for reg in regions):
            return PiecewiseVectorField("Union", regions, tolerance=tolerance)
        else:
            return None

    def __str__(self):
        return f"<{type(self).__name__} {self.name}>"


class PolygonalVectorField(VectorField):
    """A piecewise-constant vector field defined over polygonal cells.

    Arguments:
        name (str): name for debugging.
        cells: a sequence of cells, with each cell being a pair consisting of a Shapely
            geometry and a heading. If the heading is :obj:`None`, we call the given
            **headingFunction** for points in the cell instead.
        headingFunction: function computing the heading for points in cells without
            specified headings, if any (default :obj:`None`).
        defaultHeading: heading for points not contained in any cell (default
            :obj:`None`, meaning reject such points).
    """

    def __init__(self, name, cells, headingFunction=None, defaultHeading=None):
        self.cells = tuple(cells)
        if headingFunction is None and defaultHeading is not None:
            headingFunction = lambda pos: defaultHeading
        self.headingFunction = headingFunction
        for cell, heading in self.cells:
            if heading is None and headingFunction is None and defaultHeading is None:
                raise RuntimeError(f"missing heading for cell of PolygonalVectorField")
        self.defaultHeading = defaultHeading
        self.rtree = shapely.STRtree([cell[0] for cell in self.cells])
        super().__init__(name, self.valueAt)

    def valueAt(self, pos):
        point = makeShapelyPoint(pos)
        candidates = self.rtree.query(point, predicate="intersects")
        if len(candidates) > 0:
            first = candidates.min()
            cell, heading = self.cells[first]
            return self.headingFunction(pos) if heading is None else heading
        if self.defaultHeading is not None:
            return self.defaultHeading
        raise RejectionException(f"evaluated PolygonalVectorField at undefined point")


class PiecewiseVectorField(VectorField):
    """A vector field defined by patching together several regions.

    The heading at a point is determined by checking each region in turn to see if it has
    an orientation and contains the point, returning the corresponding heading if so. If
    we get through all the regions, and **tolerance** is nonzero, we try again, this time
    allowing the point to be up to **tolerance** away from each region. If we still fail
    to find a region "containing" the point, then we return the **defaultHeading**, if
    any, and otherwise reject the scene.

    Arguments:
        name (str): name for debugging.
        regions (sequence of `Region` objects): the regions making up the field.
        tolerance (float): maximum distance at which to consider a point as being
            in one of the regions, if it is not otherwise contained (default 0).
        defaultHeading (float): the heading for points not in any region with an
            orientation (default :obj:`None`, meaning reject such points).
    """

    def __init__(self, name, regions, tolerance=0, defaultHeading=None):
        self.regions = tuple(region for region in regions if region.orientation)
        self.tolerance = tolerance
        self.defaultHeading = defaultHeading
        super().__init__(name, self.valueAt)

    def valueAt(self, point):
        for region in self.regions:
            if region.containsPoint(point):
                return region.orientation[point]
        if self.tolerance > 0:
            for region in self.regions:
                if region.distanceTo(point) <= self.tolerance:
                    return region.orientation[point]
        if self.defaultHeading is not None:
            return self.defaultHeading
        raise RejectionException(f"evaluated PiecewiseVectorField at undefined point")


class PolyhedronVectorField(VectorField):
    pass
